﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Contracts;
using System.IO;
using System.Linq;
using System.Net.Http;
using System.Text.RegularExpressions;
using System.Threading;
using System.Threading.Tasks;
using ClientServer.Controllers.Core.Exercises;
using ClientServer.Db;
using ClientServer.Helpers;
using ClientServer.Models;
using ClientServer.Models.CustomProjects;
using ClientServer.Models.Exercises;
using ClientServer.Models.Exercises.AfterSolutions;
using ClientServer.Models.Exercises.Release;
using ClientServer.Models.Exercises.Solution;
using Microsoft.AspNetCore.Mvc;
using Microsoft.AspNetCore.Http;
using Microsoft.EntityFrameworkCore;
using ClientServer.Models.Exercises.Tests;
using ClientServer.Models.Interfaces;
using ClientServer.Models.Users;

namespace ClientServer.Controllers.Core.Testing
{

    /*
    all manual started tests (also by tutors) should be run on the normal test server
    this way we can ensure that the submit test server has maximum capacity
    */

        
    /*
     * when an exercise is locked we block the apis that would save test results to the db
     * normal/custom test
     * we allow running as tutor because this not saves the test result
     * we allow after editor runs because they save in a different table/class
     */

    /// <summary>
    /// controller to communicate with the test server (and frontend)
    /// </summary>
    [Route(Constants.ApiPrefix + "testing")]
    public class TestingController : ControllerWithDb
    {

        //--- notes

        //maybe instead of getting all tests separately get all tests and then filter in memory?
        //bad because normally we only run 1 tests
        //if a tests has a big asset then we might load it and don't need it

        //we now get all tests with one query... however we might swallow not found tests ... that's ok maybe someone deleted this test in the meantime

        //--- end notes


        /// <summary>
        /// a unique id for the testing request (use unchecked mode in case we reach max)
        /// also use the userId as prefix in case we spread across several backend nodes
        /// BUT there is still a small chance that one user runs > 1 tests with the same userId_requestSequentialNumber
        ///
        /// this also needs to be thread safe!!
        /// </summary>
        private static long requestSequentialNumber = 0;
        //private static string TestServerApiUrl = "http://localhost:8001/api.php";
        //private static string TestServerApiUrl = "http://192.168.30.57:8001/api.php";
        //private static string TestServerApiUrl = "http://192.168.30.56:8001/api.php";
        //192.168.145.128
        //private static string TestServerApiUrl = "http://192.168.145.128:8001/api.php";


        public TestingController(YapexDbContext context) : base(context)
        {
        }


        /// <summary>
        /// compiles the given single file on the test server
        /// user can check file that is not yet saved
        /// THIS only checks if the user is logged in, no time is up check for some exercise or release
        ///
        /// files are used from frontend --> readonly replace is not need (replace solution readonly file with exercise template readonly file)
        /// </summary>
        /// <param name="runCompileSingleFileCommand">the command (with all dependencies)</param>
        [HttpPost("compile/do/exercise")]
        public async Task RequestCompileSingleFile(
            [FromBody] RunCompileSingleFileCommandForBackend runCompileSingleFileCommand)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            PLang pLangToUse = null;
            pLangToUse = await _context.PLangs.FirstOrDefaultAsync(p => p.Id == runCompileSingleFileCommand.PlangId);

            if (pLangToUse == null)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "p lang not found")));
                return;
            }

            var systemSettings = await _context.SystemSettings.FirstOrDefaultAsync();

            if (systemSettings == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "system settings not found")));
                return;
            }

            var participation = await _context.ExerciseReleaseWithUserAsParticipations
                .Include(p => p.ExerciseRelease)
                .FirstOrDefaultAsync(p => p.UserId == userId
                                     //&& p.ExerciseRelease.IsReleased  //compiling is ok
                                     && p.ExerciseRelease.GeneratedCode == runCompileSingleFileCommand.GeneratedCode);

            if (participation?.ExerciseRelease == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no participation or release found")));
                return;
            }

            //compiling is ok
            //if (ReleaseHelper.IsTimeUp(participation, participation.ExerciseRelease, this)) return;


            if (runCompileSingleFileCommand.Files == null || runCompileSingleFileCommand.Files.Count == 0)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "no files found")));
                return;
            }

            var files = new List<SolutionFile>();
            SolutionFile mainFile = null;

            for (int i = 0; i < runCompileSingleFileCommand.Files.Count; i++)
            {
                var solutionFileForTestServer = runCompileSingleFileCommand.Files[i];
                var solutionFile = new SolutionFile()
                {
                    Id = i,
                    Content = solutionFileForTestServer.FileContent,
                    FileNameWithExtension = solutionFileForTestServer.FileName,
                };
                files.Add(solutionFile);

                if (runCompileSingleFileCommand.MainFileName == solutionFile.FileNameWithExtension)
                {
                    mainFile = solutionFile;
                }
            }

            if (mainFile == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "could not find main file")));
                return;
            }

            var solution = new Solution()
            {
                CreatedAt = DateTime.Now,
                Note = "compile single file",
                SolutionFiles = files,
                MainFile = mainFile,
                MainFileId = mainFile.Id,
            };

            var test = new AllTest
            {
                TestType = new TestType()
                {
                    DisplayName = "",
                    InternalName = Constants.CompileTestInternalName
                },
                TestSettings = new TestSettings()
                {
                    Id = -1,
                    TimeoutInMs = systemSettings.JustRunProgramTimeoutInMs,
                    MemoryLimitInKb = systemSettings.JustRunProgramMemoryLimitInKb,
                    MaxDiskSpaceInKb = systemSettings.JustRunProgramMaxDiskSpaceInKb,
                    CompileTimeoutInMs = systemSettings.JustRunProgramCompileTimeoutInMs,
                    CompilerOptions = "",
                }
            };

            #region check invalid file names on test assets and solution files

            //no test assets here
            if (!(await ValidateFileNames(solution.SolutionFiles, new List<ITestAsset>()))) return;

            #endregion

            Tuple<ResponseFromTestServer, string> answerFromTestServer = null;

            try
            {
                //use the running userid as request distinct number
                answerFromTestServer = await this._RunTestOnTestServer(solution, new List<AllTest>() {test},
                    pLangToUse, systemSettings.TestServerUrl, false, userId, systemSettings);
            }
            catch (Exception ex)
            {
                Console.WriteLine("ERROR: test server timeout hit?: " + ex);
                //probably timeout hit (not from the test program but the request
                string reason = "";
                if (AppConfiguration.IsDebugMode)
                {
                    reason = ", error: " + ex;
                }

                await
                    Response.WriteAsync(
                        Jc.Serialize(
                            new BasicResponse(ResponseCode.ServerError, "could not reach test server" + reason)));
                return;
            }


            await _WriteCompileTestResult(answerFromTestServer);
        }


        /// <summary>
        /// compiles the given single file on the test server
        /// THIS only checks if the user is logged in, no time is up check for some exercise or release
        ///
        /// files are used from frontend --> readonly replace is not need (replace solution readonly file with exercise template readonly file)
        /// </summary>
        /// <param name="runCompileSingleFileCommand">the command (with all dependencies)</param>
        [HttpPost("compile/custom/project")]
        public async Task RequestCompileSingleFileCustomProject(
            [FromBody] RunCompileSingleFileCommandCustomProjectForBackend runCompileSingleFileCommand)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            PLang pLangToUse = null;
            pLangToUse = await _context.PLangs.FirstOrDefaultAsync(p => p.Id == runCompileSingleFileCommand.PlangId);

            if (pLangToUse == null)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "p lang not found")));
                return;
            }

            var systemSettings = await _context.SystemSettings.FirstOrDefaultAsync();

            if (systemSettings == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "system settings not found")));
                return;
            }

            var customProject = await _context.CustomProjects
                    .FirstOrDefaultAsync(p => p.Id == runCompileSingleFileCommand.CustomProjectId && p.UserId == userId)
                ;

            if (customProject == null)
            {
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project not found")));
                return;
            }

            if (runCompileSingleFileCommand.Files == null || runCompileSingleFileCommand.Files.Count == 0)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "no files found")));
                return;
            }

            var files = new List<SolutionFile>();
            SolutionFile mainFile = null;

            for (int i = 0; i < runCompileSingleFileCommand.Files.Count; i++)
            {
                var solutionFileForTestServer = runCompileSingleFileCommand.Files[i];
                var solutionFile = new SolutionFile()
                {
                    Id = i,
                    Content = solutionFileForTestServer.FileContent,
                    FileNameWithExtension = solutionFileForTestServer.FileName,
                };
                files.Add(solutionFile);

                if (runCompileSingleFileCommand.MainFileName == solutionFile.FileNameWithExtension)
                {
                    mainFile = solutionFile;
                }
            }

            if (mainFile == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "could not find main file")));
                return;
            }

            var solution = new Solution()
            {
                CreatedAt = DateTime.Now,
                Note = "compile single file",
                SolutionFiles = files,
                MainFile = mainFile,
                MainFileId = mainFile.Id,
            };

            var test = new AllTest()
            {
                TestType = new TestType()
                {
                    DisplayName = "",
                    InternalName = Constants.CompileTestInternalName
                },
                TestSettings = new TestSettings()
                {
                    Id = -1,
                    TimeoutInMs = systemSettings.JustRunProgramTimeoutInMs,
                    MemoryLimitInKb = systemSettings.JustRunProgramMemoryLimitInKb,
                    MaxDiskSpaceInKb = systemSettings.JustRunProgramMaxDiskSpaceInKb,
                    CompileTimeoutInMs = systemSettings.JustRunProgramCompileTimeoutInMs,
                    CompilerOptions = "",
                }
            };

            #region check invalid file names on test assets and solution files

            //no test assets here
            if (!(await ValidateFileNames(solution.SolutionFiles, new List<ITestAsset>()))) return;

            #endregion

            Tuple<ResponseFromTestServer, string> answerFromTestServer = null;

            try
            {
                //use the running userid as request distinct number
                answerFromTestServer = await this._RunTestOnTestServer(solution, new List<AllTest>() {test},
                    pLangToUse, systemSettings.TestServerUrl, false, userId, systemSettings);
            }
            catch (Exception ex)
            {
                Console.WriteLine("ERROR: test server timeout hit?: " + ex);
                //probably timeout hit (not from the test program but the request
                string reason = "";
                if (AppConfiguration.IsDebugMode)
                {
                    reason = ", error: " + ex;
                }

                await
                    Response.WriteAsync(
                        Jc.Serialize(
                            new BasicResponse(ResponseCode.ServerError, "could not reach test server" + reason)));
                return;
            }

            await _WriteCompileTestResult(answerFromTestServer);
        }

        /// <summary>
        /// compiles the given single file on the test server
        /// THIS checks if the user has assess permission in the group connected to the exercise (managing group)
        /// </summary>
        /// <param name="runCompileSingleFileCommand">the command (with all dependencies)</param>
        [HttpPost("compile/tutor/view")]
        public async Task RequestCompileSingleFileTutorView([FromBody] RunCompileSingleFileCommandForBackend runCompileSingleFileCommand)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            if (string.IsNullOrWhiteSpace(runCompileSingleFileCommand.GeneratedCode))
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release code not found")));
                return;
            }

            var systemSettings = await _context.SystemSettings.FirstOrDefaultAsync();

            if (systemSettings == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "system settings not found")));
                return;
            }
            
            //check if user is really a tutor (has permission)
            var exerciseRelease =
                await _context.ExerciseReleases.FirstOrDefaultAsync(p => p.GeneratedCode == runCompileSingleFileCommand.GeneratedCode);

            if (exerciseRelease == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no release found")));
                return;
            }

            var userGroupId = await _context.ExerciseReleases
                .Where(p => p.GeneratedCode == runCompileSingleFileCommand.GeneratedCode)
                .Select(p => p.Exercise.UserGroupId)
                .FirstOrDefaultAsync();

            if (userGroupId <= 0)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "user group not found")));
                return;
            }

            if (await base.HasGroupPermission(userGroupId, (groupPermissions) => groupPermissions.CanAssessExercises) ==
                false)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no permission")));
                return;
            }


            PLang pLangToUse = null;
            pLangToUse = await _context.PLangs.FirstOrDefaultAsync(p => p.Id == runCompileSingleFileCommand.PlangId);

            if (pLangToUse == null)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "p lang not found")));
                return;
            }

             if (runCompileSingleFileCommand.Files == null || runCompileSingleFileCommand.Files.Count == 0)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "no files found")));
                return;
            }

            var files = new List<SolutionFile>();
            SolutionFile mainFile = null;

            for (int i = 0; i < runCompileSingleFileCommand.Files.Count; i++)
            {
                var solutionFileForTestServer = runCompileSingleFileCommand.Files[i];
                var solutionFile = new SolutionFile()
                {
                    Id = i,
                    Content = solutionFileForTestServer.FileContent,
                    FileNameWithExtension = solutionFileForTestServer.FileName,
                };
                files.Add(solutionFile);

                if (runCompileSingleFileCommand.MainFileName == solutionFile.FileNameWithExtension)
                {
                    mainFile = solutionFile;
                }
            }

            if (mainFile == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "could not find main file")));
                return;
            }

            var solution = new Solution()
            {
                CreatedAt = DateTime.Now,
                Note = "compile single file",
                SolutionFiles = files,
                MainFile = mainFile,
                MainFileId = mainFile.Id,
            };

            var test = new AllTest()
            {
                TestType = new TestType()
                {
                    DisplayName = "",
                    InternalName = Constants.CompileTestInternalName
                },
                TestSettings = new TestSettings()
                {
                    Id = -1,
                    TimeoutInMs = systemSettings.JustRunProgramTimeoutInMs,
                    MemoryLimitInKb = systemSettings.JustRunProgramMemoryLimitInKb,
                    MaxDiskSpaceInKb = systemSettings.JustRunProgramMaxDiskSpaceInKb,
                    CompileTimeoutInMs = systemSettings.JustRunProgramCompileTimeoutInMs,
                    CompilerOptions = "",
                }
            };


            #region check invalid file names on test assets and solution files

            //no test assets here
            if (!(await ValidateFileNames(solution.SolutionFiles, new List<ITestAsset>()))) return;

            #endregion

            Tuple<ResponseFromTestServer, string> answerFromTestServer = null;

            try
            {
                //use the running userid as request distinct number
                answerFromTestServer = await this._RunTestOnTestServer(solution, new List<AllTest>() {test},
                    pLangToUse, systemSettings.TestServerUrl, false, userId, systemSettings);
            }
            catch (Exception ex)
            {
                Console.WriteLine("ERROR: test server timeout hit?: " + ex);
                //probably timeout hit (not from the test program but the request
                string reason = "";
                if (AppConfiguration.IsDebugMode)
                {
                    reason = ", error: " + ex;
                }

                await
                    Response.WriteAsync(
                        Jc.Serialize(
                            new BasicResponse(ResponseCode.ServerError, "could not reach test server" + reason)));
                return;
            }


            await _WriteCompileTestResult(answerFromTestServer);
        }

        private async Task _WriteCompileTestResult(Tuple<ResponseFromTestServer, string> answerFromTestServer)
        {
            if (answerFromTestServer?.Item1 == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError,
                            answerFromTestServer?.Item2)));
                return;
            }

            //check if answer is correct

            if (answerFromTestServer.Item1.ResponseCode == (int) TestServerResponseCode.Ok)
            {
                //all ok but test could have failed
            }
            else if (answerFromTestServer.Item1.ResponseCode == (int) TestServerResponseCode.MaxParallelTestsReached)
            {
                //send the user answer to retry later
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.RetryLater, "")));
                return;
            }
            else
            {
                //some other error
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError, "test server error code: "
                                                                                 +
                                                                                 answerFromTestServer.Item1
                                                                                     .ResponseCode + " message: " +
                                                                                 answerFromTestServer.Item1
                                                                                     .Message)));
                return;
            }


            //safety check
            if (answerFromTestServer.Item1?.Data == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError, "invalid response from test server")));
                return;
            }

            List<AnswerTestFromTestServer> testResults = answerFromTestServer.Item1.Data;

            if (testResults.Count == 0)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError, "invalid response from test server")));
                return;
            }

            AnswerTestFromTestServer testResult = testResults[0];

            //in case the test-runner does not restrict this...
            bool characterLimitExceeded = testResult.Protocol.Length > YapexDbContext.TestProtocolMaxStringLength;

            if (characterLimitExceeded)
            {
                //cut last part
                testResult.Protocol = testResult.Protocol.Substring(0, YapexDbContext.TestProtocolMaxStringLength);
                testResult.CharacterLimitExceeded = true;
            }
                                           
            string[] protocol =
                testResult.Protocol.Split(new string[] {Constants.TestProtocolSplitString},
                    StringSplitOptions.RemoveEmptyEntries);

            //and give user feedback
            var response = new TestAnswerFromBackend
            {
                TestId = testResult.TestId,
                Passed = testResult.Passed,
                HasCompiled = testResult.HasCompiled,
                ProgramExitCode = testResult.ProgramExitCode,
                TestResultCode = testResult.TestResultCode,
                Protocol = protocol.ToList(),
                TestServerCode = testResult.TestServerCode,
                TestServerMessage = testResult.TestServerMessage,
                RunnerVersion = testResult.RunnerVersion,
                TestServerVersion= testResult.TestServerVersion,
                TimeForCompiling = testResult.TimeForCompiling,
                TimeForUserProgram = testResult.TimeForUserProgram,
                TimeoutInMsUsed = testResult.TimeoutInMsUsed,
                CompileTimeoutInMsUsed = testResult.CompileTimeoutInMsUsed,
                CharacterLimitExceeded = testResult.CharacterLimitExceeded,
                CharacterLimitUsed = YapexDbContext.TestProtocolMaxStringLength
            };


            await Response.WriteAsync(
                Jc.Serialize(new BasicResponseWithData<TestAnswerFromBackend>(ResponseCode.Ok, "compile file finished",
                    response)));
        }

        /// <summary>
        /// just runs a program
        /// </summary>
        /// <param name="command"></param>
        [HttpPost("justrunprogram/normal")]
        public async Task RequestJustRunProgram([FromBody] JustRunProgramTestCommandForBackend command)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            if (string.IsNullOrWhiteSpace(command.ReleaseCode))
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise not found")));
                return;
            }

            var systemSettings = await _context.SystemSettings.FirstOrDefaultAsync();

            if (systemSettings == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "system settings not found")));
                return;
            }

            var participation = await _context.ExerciseReleaseWithUserAsParticipations
                .Include(p => p.ExerciseRelease)
                .ThenInclude(p => p.Exercise.CodeTemplates)
                .ThenInclude(p => p.TemplateFiles)
                .FirstOrDefaultAsync(p => p.UserId == userId
                                     //&& p.ExerciseRelease.IsReleased // is ReleaseHelper.IsTimeUp checks for is released so wo do not need to doulbe check this
                                     && p.ExerciseRelease.GeneratedCode == command.ReleaseCode);


            if (participation?.ExerciseRelease == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no participation or release found")));
                return;
            }


            //we need to allow just run (for the after solution)
            //if (ReleaseHelper.IsTimeUp(participation, participation.ExerciseRelease, this)) return;


            //then we can get the plang directly from the user solution...
            PLang pLangToUse = await _context.PLangs
                .FirstOrDefaultAsync(p => p.Id == command.PlangId);

            if (pLangToUse == null)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "p lang not found")));
                return;
            }

            var exerciseRelease = participation.ExerciseRelease;

            //forge temp test

            string testContent = "";

            if (string.IsNullOrWhiteSpace(command.CmdArgs) == false)
            {
                if (command.CmdArgs.Length > Constants.MaxJustRunProgramArgsLength) {
                    command.CmdArgs = "";
                }

                testContent = Constants.TestProtocolArgsString + " " + command.CmdArgs;
            }

            AllTest justRunTest = this._CreateJustRunTest(testContent, systemSettings);

            var allTests = new List<AllTest>();
            allTests.Add(justRunTest);

            List<TestAnswerFromBackend> results;

            #region get solution

            //we need to distinguish two cases
            
            //release is already finished and the user edits the subsequent editor
            //or a normal solution edit (solution can still be edited)

            Solution userSolution;
            
            //check if available working time is up
            if (ReleaseHelper.IsTimeUp(participation, participation.ExerciseRelease))
            {
                //then the user is editing the after solution...
                
                var afterSolution = await _context.AfterSolutions
                    .Include(p => p.SolutionFiles)
                    .ThenInclude(p => p.TemplateFile)
                    .Include(p => p.TestResults) //only include tests here because its a normal test (not a custom test)
                    .FirstOrDefaultAsync(
                        p => p.SolutionUserId == userId && p.SolutionExerciseReleaseId == exerciseRelease.Id && p.SolutionPLangId == pLangToUse.Id);
                
                if (afterSolution == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "user after solution not found")));
                    return;
                }
                
                userSolution = new Solution()
                {
                    MainFileId = afterSolution.MainFileId,
                    PLangId = afterSolution.SolutionPLangId, //to get code templates
                    SolutionFiles = afterSolution.SolutionFiles.Select(p => new SolutionFile()
                    {
                        Id = p.Id,
                        Content = p.Content,
                        DisplayIndex = p.DisplayIndex,
                        IsDisplayed = p.IsDisplayed,
                        Solution = null,
                        FileNameWithExtension = p.FileNameWithExtension,
                        SolutionPLangId = afterSolution.SolutionPLangId,
                        SolutionExerciseReleaseId = afterSolution.SolutionExerciseReleaseId,
                        SolutionUserId = afterSolution.SolutionUserId,
                        TemplateFileId = p.TemplateFileId,
                        TemplateFile = null
                    }).ToList()
                };

                userSolution.MainFile = userSolution.SolutionFiles.FirstOrDefault(p => p.Id == userSolution.MainFileId);

                if (userSolution.MainFile == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "after solution main file not found")));
                    return;
                }
            }
            else
            {
               //normal edit
                
                userSolution = await _context.Solutions
                    .Include(p => p.SolutionFiles)
                    .ThenInclude(p => p.TemplateFile)
                    .Include(p => p.TestResults) //only include tests here because its a normal test (not a custom test)
                    .FirstOrDefaultAsync(
                        p => p.UserId == userId && p.ExerciseReleaseId == exerciseRelease.Id && p.PLangId == pLangToUse.Id);


                if (userSolution == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "user solution not found")));
                    return;
                }
            }
            
            

            var codeTemplate = exerciseRelease.Exercise.CodeTemplates
                    .FirstOrDefault(p => p.PLangId == userSolution.PLangId)
                ;
            
            if (codeTemplate == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "codeTemplate not found")));
                return;
            }
            
            //make readonly files up-to-date
            var errorMsg = UserSolutionHelper.ReplaceReadonlyFiles(userSolution, codeTemplate, false, true);

            if (errorMsg != null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError, errorMsg)));
                return;
            }

            //--- run test

           
            results = await this._RunTempTests(userId, userSolution, pLangToUse, allTests, systemSettings, false);

            //this is handled in the method
            if (results == null) return;



            #endregion

            

            TestAnswerFromBackend result = results[0];

            await Response.WriteAsync(
                Jc.Serialize(new BasicResponseWithData<TestAnswerFromBackend>(ResponseCode.Ok, "",
                    result)));
        }

        /// <summary>
        /// just runs a program
        /// </summary>
        /// <param name="command"></param>
        [HttpPost("justrunprogram/custom/project")]
        public async Task RequestJustRunProgram([FromBody] JustRunProgramTestCustomProjectCommandForBackend command)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var systemSettings = await _context.SystemSettings.FirstOrDefaultAsync();

            if (systemSettings == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "system settings not found")));
                return;
            }

            var customProject = await _context.CustomProjects
                    .FirstOrDefaultAsync(p => p.Id == command.CustomProjectId && p.UserId == userId)
                ;

            //then we can get the plang directly from the user solution...
            PLang pLangToUse = null;
            pLangToUse = await _context.PLangs
                .FirstOrDefaultAsync(p => p.Id == command.PlangId);

            if (pLangToUse == null)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "p lang not found")));
                return;
            }

            string testContent = "";

            if (string.IsNullOrWhiteSpace(command.CmdArgs) == false)
            {
                if (command.CmdArgs.Length > Constants.MaxJustRunProgramArgsLength) {
                    command.CmdArgs = "";
                }
                
                testContent = Constants.TestProtocolArgsString + " " + command.CmdArgs;
            }

            //forge temp test

            AllTest justRunTest = this._CreateJustRunTest(testContent, systemSettings);


            //get solution

            //do not include solution files here ... it might be too many data to load...
            //just find the correct version then start another query to get the solution version + files


            var solution = await _context.CustomProjectSolutions
                    .Include(p => p.SolutionFiles)
                    .Include(p => p.TestResults)
                    .FirstOrDefaultAsync(p => p.CustomProjectId == customProject.Id && p.PLangId == pLangToUse.Id)
                ;

            if (solution == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "solution not found")));
                return;
            }


            var solutionFiles = solution.SolutionFiles.Select(p => new SolutionFile()
            {
                Id = p.Id,
                Content = p.Content,
                DisplayIndex = p.DisplayIndex,
                FileNameWithExtension = p.FileNameWithExtension,
                IsDisplayed = p.IsDisplayed,
                TemplateFile = null,
                TemplateFileId = -1
            }).ToList();

            var mainFile = solutionFiles.FirstOrDefault(p => p.Id == solution.MainFileId);

            if (mainFile == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "main file not found")));
                return;
            }

            var solutionVersion = new Solution()
            {
                CreatedAt = solution.CreatedAt,
                SolutionFiles = solutionFiles,
                CustomTestResults = null,
                Note = "",
                TestResults = null,
                MainFileId = solution.MainFileId,
                MainFile = mainFile
            };


            //--- run test

            var allTests = new List<AllTest>();
            allTests.Add(justRunTest);

            List<TestAnswerFromBackend> results =
                await this._RunTempTests(userId, solutionVersion, pLangToUse, allTests, systemSettings, false);

            //this is handled in the method
            if (results == null) return;


            TestAnswerFromBackend result = results[0];

            await Response.WriteAsync(
                Jc.Serialize(new BasicResponseWithData<TestAnswerFromBackend>(ResponseCode.Ok, "",
                    result)));
        }

         /// <summary>
        /// just runs the given program
        /// because the tutor changes are only temporarily we need to files from frontend...
        /// THIS checks if the user has assess permission in the group connected to the exercise (managing group)
        /// </summary>
        /// <param name="command">the command (with all dependencies)</param>
        [HttpPost("justrunprogram/tutor/view")]
        public async Task RequestJustRunProgramTutorView([FromBody] JustRunProgramTestAsTutorCommandForBackend command)
        {
            //TODO most of the code is copied from RequestCompileSingleFileAsTutor

            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            if (string.IsNullOrWhiteSpace(command.ReleaseCode))
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release code not found")));
                return;
            }

            var systemSettings = await _context.SystemSettings.FirstOrDefaultAsync();

            if (systemSettings == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "system settings not found")));
                return;
            }
            
            //check if user is really a tutor (has permission)
            var exerciseRelease =
                await _context.ExerciseReleases.FirstOrDefaultAsync(p => p.GeneratedCode == command.ReleaseCode);

            if (exerciseRelease == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no release found")));
                return;
            }

            var userGroupId = await _context.ExerciseReleases
                .Where(p => p.GeneratedCode == command.ReleaseCode)
                .Select(p => p.Exercise.UserGroupId)
                .FirstOrDefaultAsync();

            if (userGroupId <= 0)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "user group not found")));
                return;
            }

            if (await base.HasGroupPermission(userGroupId, (groupPermissions) => groupPermissions.CanAssessExercises) ==
                false)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no permission")));
                return;
            }


            PLang pLangToUse = null;
            pLangToUse = await _context.PLangs.FirstOrDefaultAsync(p => p.Id == command.PlangId);

            if (pLangToUse == null)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "p lang not found")));
                return;
            }

             if (command.Files == null || command.Files.Count == 0)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "no files found")));
                return;
            }

            var files = new List<SolutionFile>();
            SolutionFile mainFile = null;

            for (int i = 0; i < command.Files.Count; i++)
            {
                var solutionFileForTestServer = command.Files[i];
                var solutionFile = new SolutionFile()
                {
                    Id = i,
                    Content = solutionFileForTestServer.FileContent,
                    FileNameWithExtension = solutionFileForTestServer.FileName,
                };
                files.Add(solutionFile);

                if (command.MainFileName == solutionFile.FileNameWithExtension)
                {
                    mainFile = solutionFile;
                }
            }

            if (mainFile == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "could not find main file")));
                return;
            }

            var solution = new Solution()
            {
                CreatedAt = DateTime.Now,
                Note = "compile single file",
                SolutionFiles = files,
                MainFile = mainFile,
                MainFileId = mainFile.Id,
            };

            #region check invalid file names on test assets and solution files

            //no test assets here
            if (!(await ValidateFileNames(solution.SolutionFiles, new List<ITestAsset>()))) return;

            #endregion

            //forge temp test

            string testContent = "";

            if (string.IsNullOrWhiteSpace(command.CmdArgs) == false)
            {
                if (command.CmdArgs.Length > Constants.MaxJustRunProgramArgsLength) {
                    command.CmdArgs = "";
                }

                testContent = Constants.TestProtocolArgsString + " " + command.CmdArgs;
            }

            AllTest justRunTest = this._CreateJustRunTest(testContent, systemSettings);

             //--- run test

            var allTests = new List<AllTest>();
            allTests.Add(justRunTest);

            List<TestAnswerFromBackend> results =
                await this._RunTempTests(userId, solution, pLangToUse, allTests, systemSettings, false);

            //this is handled in the method
            if (results == null) return;


            TestAnswerFromBackend result = results[0];

            await Response.WriteAsync(
                Jc.Serialize(new BasicResponseWithData<TestAnswerFromBackend>(ResponseCode.Ok, "",
                    result)));
        }

        private AllTest _CreateJustRunTest(string testContent, SystemSetting systemSettings)
        {
            var testSettings = new TestSettings()
            {
                Id = -1,
                TimeoutInMs = systemSettings.JustRunProgramTimeoutInMs,
                MemoryLimitInKb = systemSettings.JustRunProgramMemoryLimitInKb,
                MaxDiskSpaceInKb = systemSettings.JustRunProgramMaxDiskSpaceInKb,
                CompileTimeoutInMs = systemSettings.JustRunProgramCompileTimeoutInMs,
                CompilerOptions = "",
            };
            
            var justRunTest = new AllTest()
            {
                Id = -1,
                DisplayName = "Just Run Program",
                TestAssetsWithContent = new List<AllTestAsset>(),
                TestSettings = testSettings,
                TestType = new TestType()
                {
                    Id = -1,
                    DisplayName = "Just Run",
                    InternalName = "justRunTest"
                },
                Content = testContent,
            };

            return justRunTest;
        }


        //--- tests where we store the result

        /// <summary>
        /// runs a test
        /// </summary>
        /// <param name="command">the test to run</param>
        [HttpPost("run")]
        public async Task RequestRunTest([FromBody] RunTestCommandForBackend command)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            if (string.IsNullOrWhiteSpace(command.GeneratedCode))
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise not found")));
                return;
            }

            var systemSettings = await _context.SystemSettings.FirstOrDefaultAsync();

            if (systemSettings == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "system settings not found")));
                return;
            }

            var participation = await _context.ExerciseReleaseWithUserAsParticipations
                .Include(p => p.ExerciseRelease)
                .ThenInclude(p => p.Exercise.CodeTemplates)
                .ThenInclude(p => p.TemplateFiles)
                .FirstOrDefaultAsync(p => p.UserId == userId
                                     //&& p.ExerciseRelease.IsReleased // is ReleaseHelper.IsTimeUp checks for is released so wo do not need to double check this
                                     && p.ExerciseRelease.GeneratedCode == command.GeneratedCode);


            if (participation?.ExerciseRelease == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no participation or release found")));
                return;
            }

            if (participation.ExerciseRelease.Exercise == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise not found")));
                return;
            }

            if (participation.ExerciseRelease.Exercise.IsPermanentlyLocked)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "exercise is permanently locked")));
                return;
            }

            if (ReleaseHelper.IsTimeUp(participation, participation.ExerciseRelease, this)) return;


            if (!command.IsCustomTest && !command.IsNormalTest &&
                !command.IsSubmitTest)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "test type not set")));
                return;
            }

            //then we can get the plang directly from the user solution...
            PLang pLangToUse = await _context.PLangs
                .FirstOrDefaultAsync(p => p.Id == command.PlangId);

            if (pLangToUse == null)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "p lang not found")));
                return;
            }

            //which test kind??
            if (command.IsNormalTest)
            {
                await _RunNormalOrSubmitTest(command, userId, pLangToUse, participation.ExerciseRelease, false, systemSettings);
            }
            else if (command.IsCustomTest)
            {
                await _RunCustomTest(command, userId, pLangToUse, participation.ExerciseRelease, systemSettings);
            }
            else if (command.IsSubmitTest)
            {
                Console.WriteLine("[WARNING] running submit tests is not supported here (yet)");
                
                //not used from frontend... user cannot run submit tests... maybe later for tutors?
//                await _RunNormalOrSubmitTest(command, userId, pLangToUse, participation.ExerciseRelease, true, systemSettings);
            }
        }


        /// <summary>
        /// validates the file names for the soltuion files and the test assets (files)
        ///
        /// TODO maybe use <see cref="Files.ValidFileName"/>??
        /// </summary>
        /// <param name="solutionFiles">the solution files</param>
        /// <param name="testAssets">the test assets to check too </param>
        /// <returns>true: all ok, false: some invalid file name, writes the response with the invalid file name</returns>
        private async Task<bool> ValidateFileNames(IEnumerable<ISolutionFile> solutionFiles,
            IEnumerable<ITestAsset> testAssets)
        {
            string regexString = "[" + Regex.Escape(new string(Path.GetInvalidFileNameChars())) + "]";
            Regex containsABadCharacter = new Regex(regexString);

            //check all file for invalid name
            ISolutionFile invalidFile =
                solutionFiles.FirstOrDefault(p => containsABadCharacter.IsMatch(p.FileNameWithExtension));

            if (invalidFile != null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponseWithData<string>(ResponseCode.InvalidRequest,
                            "invalid file name found", invalidFile.FileNameWithExtension)));
                return false;
            }

            //check for invalid asset names too
            var invalidAsset = testAssets.FirstOrDefault(p => containsABadCharacter.IsMatch(p.DisplayName));

            if (invalidAsset != null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponseWithData<string>(ResponseCode.InvalidRequest,
                            "invalid test asset name found", invalidAsset.DisplayName)));
                return false;
            }

            return true;
        }

        /// <summary>
        /// runs a normal test on the test server (called from RunTest)
        /// this does not checks the available working time or if the release is still released!
        /// </summary>
        /// <param name="runNormalTestCommand">the test to run (with all dependecies) from frontend</param>
        /// <param name="userId">the user id</param>
        /// <param name="plang">the plang to use</param>
        /// <param name="exerciseRelease">the exercise release to find the user solution, make sure exercise and coded template is loadedd!</param>
        /// <param name="isSubmitTest">true: is a submit test, false: not</param>
        /// <param name="systemSettings"></param>
        private async Task _RunNormalOrSubmitTest(RunTestCommandForBackend runNormalTestCommand, int userId,
            PLang plang, ExerciseRelease exerciseRelease, bool isSubmitTest, SystemSetting systemSettings)
        {
            if (isSubmitTest)
            {
                //the user group that manages the exercise
                var targetUserGroup =
                    await _context.Exercises.Where(p => p.Id == exerciseRelease.ExerciseId)
                        .Select(p => p.UserGroupId)
                        .FirstOrDefaultAsync();

                //test special permission to execute a submit test
                if (!await base.HasGroupPermission(targetUserGroup,
                    permission => permission != null && permission.CanAssessExercises))
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                    return;
                }

                //submit test is EVERY time run after the release is finished... not directly by a user 

                //so only check release
                if (exerciseRelease.IsReleased == true)
                {
                    await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.ReleaseStillActive, "")));
                    return;
                }
            }
            else
            {
                //this is already done in the calling method (where the http entry point is...)

                //if it's a normal test then check working time (the submission assessment worker do not use this method so it's ensured that
                //this is run by a user (wich should have a participation)
                //var participation = _context.ExerciseReleaseWithUserAsParticipations
                //    .FirstOrDefault(p => p.ExerciseRelease.GeneratedCode == exerciseRelease.GeneratedCode && p.UserId == userId);

                //if (participation == null)
                //{
                //    return;
                //}

                //check if available working time is up

                //check if available working time is up
                //if (ReleaseHelper.IsTimeUp(participation, exerciseRelease, this))
                //{
                //    return;
                //}
            }

            //if we requested more... see tooManyTestsRequested

            var testIds = runNormalTestCommand.TestIds.Take(isSubmitTest
                    ? systemSettings.MaxNumberOfTestsWithOneRequestSubmitTestServer
                    : systemSettings.MaxNumberOfTestsWithOneRequest)
                .ToList();
            ;

            var tests = await _context.Tests
                    .Include(p => p.TestSettings)
                    .Include(p => p.TestType)
                    .Include(p => p.AssetReferences)
                    .ThenInclude(p => p.FileReferenceTestAsset)
                    .Where(p => p.IsSubmitTest == isSubmitTest && p.ExerciseId == exerciseRelease.ExerciseId &&
                                testIds.Contains(p.Id))
                    .ToListAsync()
                ;

            var tooManyTestsRequested = tests.Count != runNormalTestCommand.TestIds.Count;
            
            
            //get test asset contents
            var testFilesIdsToRead = tests.SelectMany(p => p.AssetReferences)
                .Select(p => p.FileReferenceTestAssetId)
                .ToArray();

            var testAssetsBasePath = Files.GetUploadFilePath(UploadDirType.TestAssets);

            var testFilesDict = await Files.ReadUploadedFilesAsArray(testAssetsBasePath, testFilesIdsToRead);

            List<AllTest> allTests;

            try
            {
                allTests = tests.Select(p => AllTest.CreateFrom(p, testFilesDict)).ToList();
            }
            catch (Exception e)
            {
                //we create this exception so it should be ok to send just the message...
                Console.WriteLine("ERROR " + e);
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, e.Message)));
                return;
            }

            //do not include solution files here ... it might be too many data to load...
            //just find the correct version then start another query to get the solution version + files

            Solution userSolution = await _context.Solutions
                .Include(p => p.SolutionFiles)
                .Include(p => p.TestResults) //only include tests here because its a normal test (not a custom test)
                .FirstOrDefaultAsync(
                    p => p.UserId == userId && p.ExerciseReleaseId == exerciseRelease.Id && p.PLangId == plang.Id);


            if (userSolution == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "user solution not found")));
                return;
            }

            var codeTemplate = exerciseRelease.Exercise.CodeTemplates
                    .FirstOrDefault(p => p.PLangId == userSolution.PLangId)
                ;
            
            if (codeTemplate == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "codeTemplate not found")));
                return;
            }
            
            //make readonly files up-to-date
            var errorMsg = UserSolutionHelper.ReplaceReadonlyFiles(userSolution, codeTemplate, false, true);

            if (errorMsg != null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError, errorMsg)));
                return;
            }
            
            //---run the test

            var testResultAnswerFromBackend = await _RunTempTests(userId, userSolution, plang, allTests, systemSettings, false);

            if (testResultAnswerFromBackend == null)
            {
                //already handled in _RunTempTests
                return;
            }

            foreach (var testResult in testResultAnswerFromBackend)
            {
                var test = tests.FirstOrDefault(p => p.Id == testResult.TestId);

                if (test == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.ServerError,
                                "test not found for test result")));
                    return;
                }

                //get the old result
                var oldResult = userSolution.TestResults
                    .FirstOrDefault(
                        p => p.TestId == test.Id
                             && p.Solution == userSolution);
                //this is not a db query so the references are loaded

                //and store test result

                if (oldResult == null)
                {
                    //no old result, create new one
                    oldResult = new TestWithSingleSolutionAsTestResult()
                    {
                        Solution = userSolution,
                        Test = test,
                        Passed = testResult.Passed,
                        Protocol = string.Join(Constants.TestProtocolSplitString, testResult.Protocol),
                        HasCompiled = testResult.HasCompiled,
                        ProgramExitCode = testResult.ProgramExitCode,
                        TestResultCode = testResult.TestResultCode,
                        RunnerVersion = testResult.RunnerVersion,
                        TestServerVersion= testResult.TestServerVersion,
                        TimeForCompiling = testResult.TimeForCompiling,
                        TimeForUserProgram = testResult.TimeForUserProgram,
                        TimeoutInMsUsed = testResult.TimeoutInMsUsed,
                        CompileTimeoutInMsUsed = testResult.CompileTimeoutInMsUsed,
                        CharacterLimitExceeded = testResult.CharacterLimitExceeded,
                        CharacterLimitUsed = testResult.CharacterLimitUsed,
                    };

                    _context.TestWithSingleSolutionAsTestResult.Add(oldResult);
                }
                else
                {
                    //change old result
                    oldResult.Passed = testResult.Passed;
                    oldResult.Protocol = string.Join(Constants.TestProtocolSplitString, testResult.Protocol);
                    oldResult.HasCompiled = testResult.HasCompiled;
                    oldResult.ProgramExitCode = testResult.ProgramExitCode;
                    oldResult.TestResultCode = testResult.TestResultCode;

                    oldResult.RunnerVersion = testResult.RunnerVersion;
                    oldResult.TestServerVersion= testResult.TestServerVersion;
                    oldResult.TimeForCompiling = testResult.TimeForCompiling;
                    oldResult.TimeForUserProgram = testResult.TimeForUserProgram;
                    oldResult.TimeoutInMsUsed = testResult.TimeoutInMsUsed;
                    oldResult.CompileTimeoutInMsUsed = testResult.CompileTimeoutInMsUsed;
                   oldResult.CharacterLimitExceeded = testResult.CharacterLimitExceeded;
                   oldResult.CharacterLimitUsed = testResult.CharacterLimitUsed;
                }

                try
                {
                    await _context.SaveChangesAsync();
                }
                catch (Exception ex)
                {
                   await base.HandleDbError(ex);
                    return;
                }
            }


            //user could request too many tests to one user blocks the tests server...
            //but some tests might be run...
            if (tooManyTestsRequested)
            {
                if (isSubmitTest)
                {
                    await Response.WriteAsync(
                        Jc.Serialize(new BasicResponseWithData<List<TestAnswerFromBackend>>(ResponseCode.PartialOk,
                            "too many tests requested, max: "
                            + systemSettings.MaxNumberOfTestsWithOneRequestSubmitTestServer,
                            testResultAnswerFromBackend)));
                }
                else
                {
                    //normal tests
                    await Response.WriteAsync(
                        Jc.Serialize(new BasicResponseWithData<List<TestAnswerFromBackend>>(ResponseCode.PartialOk,
                            "too many tests requested, max: "
                            + systemSettings.MaxNumberOfTestsWithOneRequest,
                            testResultAnswerFromBackend)));
                }
            }
            else
            {
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<List<TestAnswerFromBackend>>(ResponseCode.Ok, "tests ran",
                        testResultAnswerFromBackend)));
            }
        }

        /// <summary>
        /// runs a test AFTER the use cannot longer edit the solution --> unreleased or time is up && HasAutomaticAssessmentFinished because only then the user should see the submit tests
        ///
        ///
        /// submit tests are run on the submit test server, normal tests are run on the normal test server
        /// </summary>
        /// <param name="command"></param>
        [HttpPost("run/after")]
        public async Task RequestRunAfterTest([FromBody] RunTestCommandForBackend command)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var systemSettings = await _context.SystemSettings.FirstOrDefaultAsync();

            if (systemSettings == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "system settings not found")));
                return;
            }
            
            //check if the user the user has a participation else he/she has no permission

            ExerciseReleaseWithUserAsParticipation releaseWithUserAsParticipation = await _context
                .ExerciseReleaseWithUserAsParticipations
                .Include(p => p.ExerciseRelease)
                .ThenInclude(p => p.Exercise.CodeTemplates)
                .ThenInclude(p => p.TemplateFiles)
                .FirstOrDefaultAsync(p => p.UserId == userId && p.ExerciseRelease.GeneratedCode == command.GeneratedCode);

            if (releaseWithUserAsParticipation == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission,
                            "participation not found")));
                return;
            }

            if (releaseWithUserAsParticipation?.ExerciseRelease == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                            "release not found")));
                return;
            }

            if (!command.IsCustomTest && !command.IsNormalTest &&
                !command.IsSubmitTest)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "test type not set")));
                return;
            }


            //check if available working time is up
            if (ReleaseHelper.IsTimeUp(releaseWithUserAsParticipation,
                releaseWithUserAsParticipation.ExerciseRelease))
            {
                //because the user can run submit tests this is only allowed after the release has finished & the assessment has run
            }
            else
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                            "release still active")));
                return;
            }

            if (command.IsSubmitTest)
            {
                //we know the user can't edit the solution here but submit tests are only visible after the auto assessment
                if (releaseWithUserAsParticipation.ExerciseRelease.HasAutomaticAssessmentFinished == false)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                "assessment has not finished yet")));
                    return;
                }
            }

            PLang pLangToUse = await _context.PLangs
                .FirstOrDefaultAsync(p => p.Id == command.PlangId);

            if (pLangToUse == null)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "p lang not found")));
                return;
            }

            AfterSolution afterSolution = await _context.AfterSolutions
                .Include(p => p.SolutionFiles)
                .Include(p => p.TestResults) //only include tests here because its a normal test (not a custom test)
                .Include(p => p.CustomTestResults)
                .FirstOrDefaultAsync(
                    p => p.SolutionUserId == userId &&
                         p.SolutionExerciseReleaseId == releaseWithUserAsParticipation.ExerciseRelease.Id &&
                         p.SolutionPLangId == pLangToUse.Id);


            if (afterSolution == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "after solution not found")));
                return;
            }

            var codeTemplate = releaseWithUserAsParticipation.ExerciseRelease.Exercise.CodeTemplates
                    .FirstOrDefault(p => p.PLangId == afterSolution.SolutionPLangId)
                ;
            
            if (codeTemplate == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "codeTemplate not found")));
                return;
            }
            
            //make readonly files up-to-date
            var errorMsg = UserSolutionHelper.ReplaceReadonlyFiles(afterSolution, codeTemplate, false, true);
            
            if (errorMsg != null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError, errorMsg)));
                return;
            }

            var userSolution = new Solution()
            {
                MainFileId = afterSolution.MainFileId,
                SolutionFiles = afterSolution.SolutionFiles.Select(p => new SolutionFile()
                {
                    Id = p.Id,
                    Content = p.Content,
                    DisplayIndex = p.DisplayIndex,
                    IsDisplayed = p.IsDisplayed,
                    Solution = null,
                    FileNameWithExtension = p.FileNameWithExtension,
                    SolutionPLangId = afterSolution.SolutionPLangId,
                    SolutionExerciseReleaseId = afterSolution.SolutionExerciseReleaseId,
                    SolutionUserId = afterSolution.SolutionUserId,
                    TemplateFileId = p.TemplateFileId,
                    TemplateFile = null
                }).ToList()
            };

            userSolution.MainFile = userSolution.SolutionFiles.FirstOrDefault(p => p.Id == userSolution.MainFileId);

            if (userSolution.MainFile == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "after solution main file not found")));
                return;
            }


            //which test kind??
            if (command.IsNormalTest || command.IsSubmitTest)
            {
                //user is only allowed to run on the normal test server

                var testIds = command.TestIds.Take(command.IsSubmitTest
                        ? systemSettings.MaxNumberOfTestsWithOneRequestSubmitTestServer
                        : systemSettings.MaxNumberOfTestsWithOneRequest)
                    .ToList();
                ;

                var tests = await _context.Tests
                        .Include(p => p.TestSettings)
                        .Include(p => p.TestType)
                        .Include(p => p.AssetReferences)
                        .ThenInclude(p => p.FileReferenceTestAsset)
                        .Where(p => p.IsSubmitTest == command.IsSubmitTest && p.ExerciseId ==
                                    releaseWithUserAsParticipation
                                        .ExerciseRelease.ExerciseId &&
                                    testIds.Contains(p.Id))
                        .ToListAsync()
                    ;

               
                
                var testFilesIdsToRead = tests.SelectMany(p => p.AssetReferences)
                    .Select(p => p.FileReferenceTestAssetId)
                    .ToArray();

                var testAssetsBasePath = Files.GetUploadFilePath(UploadDirType.TestAssets);

                var testFilesDict = await Files.ReadUploadedFilesAsArray(testAssetsBasePath, testFilesIdsToRead);

                List<AllTest> allTests;

                try
                {
                    allTests = tests.Select(p => AllTest.CreateFrom(p, testFilesDict)).ToList();
                }
                catch (Exception e)
                {
                    //we create this exception so it should be ok to send just the message...
                    Console.WriteLine("ERROR " + e);
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, e.Message)));
                    return;
                }
                
                var tooManyTestsRequested = tests.Count != command.TestIds.Count;

                //ignore isSubmit tests and use normal test conditions... this is to protect the submit test server from
                //user spamming BUT this might introduce inconsistent test results because submit test server can have different settings e.g. timeouts

                var testResultAnswerFromBackend =
                    await _RunTempTests(userId, userSolution, pLangToUse, allTests, systemSettings, false);


                if (testResultAnswerFromBackend == null)
                {
                    //already handled in _RunTempTests
                    return;
                }

                foreach (var testResult in testResultAnswerFromBackend)
                {
                    var test = tests.FirstOrDefault(p => p.Id == testResult.TestId);

                    if (test == null)
                    {
                        await
                            Response.WriteAsync(
                                Jc.Serialize(new BasicResponse(ResponseCode.ServerError,
                                    "test not found for test result")));
                        return;
                    }

                    //get the old result
                    var oldResult = afterSolution.TestResults
                        .FirstOrDefault(
                            p => p.TestId == test.Id
                                 && p.AfterSolution == afterSolution);
                    //this is not a db query so the references are loaded

                    //and store test result

                    if (oldResult == null)
                    {
                        //no old result, create new one
                        oldResult = new TestWithAfterSolutionAsTestResult()
                        {
                            AfterSolution = afterSolution,
                            Test = test,
                            Passed = testResult.Passed,
                            Protocol = string.Join(Constants.TestProtocolSplitString, testResult.Protocol),
                            HasCompiled = testResult.HasCompiled,
                            ProgramExitCode = testResult.ProgramExitCode,
                            TestResultCode = testResult.TestResultCode,
                            RunnerVersion = testResult.RunnerVersion,
                            TestServerVersion= testResult.TestServerVersion,
                            TimeForCompiling = testResult.TimeForCompiling,
                            TimeForUserProgram = testResult.TimeForUserProgram,
                            TimeoutInMsUsed = testResult.TimeoutInMsUsed,
                            CompileTimeoutInMsUsed = testResult.CompileTimeoutInMsUsed,
                            CharacterLimitExceeded = testResult.CharacterLimitExceeded,
                            CharacterLimitUsed = testResult.CharacterLimitUsed,
                        };

                        _context.TestWithAfterSolutionAsTestResults.Add(oldResult);
                    }
                    else
                    {
                        //change old result
                        oldResult.Passed = testResult.Passed;
                        oldResult.Protocol = string.Join(Constants.TestProtocolSplitString, testResult.Protocol);
                        oldResult.HasCompiled = testResult.HasCompiled;
                        oldResult.ProgramExitCode = testResult.ProgramExitCode;
                        oldResult.TestResultCode = testResult.TestResultCode;
                        
                        oldResult.RunnerVersion = testResult.RunnerVersion;
                        oldResult.TestServerVersion= testResult.TestServerVersion;
                        oldResult.TimeForCompiling = testResult.TimeForCompiling;
                        oldResult.TimeForUserProgram = testResult.TimeForUserProgram;
                        oldResult.TimeoutInMsUsed = testResult.TimeoutInMsUsed;
                        oldResult.CompileTimeoutInMsUsed = testResult.CompileTimeoutInMsUsed;
                       oldResult.CharacterLimitExceeded = testResult.CharacterLimitExceeded;
                       oldResult.CharacterLimitUsed = testResult.CharacterLimitUsed;
                    }

                    try
                    {
                        await _context.SaveChangesAsync();
                    }
                    catch (Exception ex)
                    {
                       await base.HandleDbError(ex);
                        return;
                    }
                }


                //user could request too many tests to one user blocks the tests server...
                //but some tests might be run...
                if (tooManyTestsRequested)
                {
                    if (false) //user can only run on the normal test server
                    {
                        await Response.WriteAsync(
                            Jc.Serialize(new BasicResponseWithData<List<TestAnswerFromBackend>>(ResponseCode.PartialOk,
                                "too many tests requested, max: "
                                + systemSettings.MaxNumberOfTestsWithOneRequestSubmitTestServer,
                                testResultAnswerFromBackend)));
                    }
                    else
                    {
                        //normal tests
                        await Response.WriteAsync(
                            Jc.Serialize(new BasicResponseWithData<List<TestAnswerFromBackend>>(ResponseCode.PartialOk,
                                "too many tests requested, max: "
                                + systemSettings.MaxNumberOfTestsWithOneRequest,
                                testResultAnswerFromBackend)));
                    }
                }
                else
                {
                    await Response.WriteAsync(
                        Jc.Serialize(new BasicResponseWithData<List<TestAnswerFromBackend>>(ResponseCode.Ok,
                            "tests ran",
                            testResultAnswerFromBackend)));
                }
            }
            else if (command.IsCustomTest)
            {
                

                DefaultCustomTestSettings defaultCustomTestSettings = await _context.DefaultCustomTestSettings
                        .FirstOrDefaultAsync(p =>
                        p.ExerciseId == releaseWithUserAsParticipation.ExerciseRelease.ExerciseId);

                if (defaultCustomTestSettings == null)
                {
                    await Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no custom test settings found")));
                    return;
                }

                //user is only allowed to run on the normal test server
                var testIds = command.TestIds.Take(systemSettings.MaxNumberOfTestsWithOneRequest)
                    .ToList();
                ;

                var tests = await _context.CustomTest
                        //.Include(p => p.TestSettings)
                        .Include(p => p.TestType)
                        .Include(p => p.AssetReferences)
                        .ThenInclude(p => p.FileReferenceUserFileAsset)
                        .Where(p =>
                            p.UserId == userId &&
                            p.ExerciseReleaseId == releaseWithUserAsParticipation.ExerciseReleaseId &&
                            testIds.Contains(p.Id))
                        .ToListAsync()
                    ;

                var tooManyTestsRequested = tests.Count != command.TestIds.Count;
                
                var fileReferenceIds = tests.SelectMany(p => p.AssetReferences)
                    .Select(p => p.FileReferenceUserFileAssetId)
                    .ToArray();
                
                var testAssetsBasePath = Files.GetUploadFilePath(UploadDirType.UserAssets);

                var testFilesDict = await Files.ReadUploadedFilesAsArray(testAssetsBasePath, fileReferenceIds);

                List<AllTest> allTests;


                try
                {
                    allTests = tests.Select(p => AllTest.CreateFrom(p, defaultCustomTestSettings, testFilesDict)).ToList();
                }
                catch (Exception e)
                {
                    //we create this exception so it should be ok to send just the message...
                    Console.WriteLine("ERROR " + e);
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, e.Message)));
                    return;
                }

                //ignore isSubmit tests and use normal test conditions... this is to protect the submit test server from
                //user spamming BUT this might introduce inconsistent test results because submit test server can have different settings e.g. timeouts

                var testResultAnswerFromBackend =
                    await _RunTempTests(userId, userSolution, pLangToUse, allTests, systemSettings, false);


                if (testResultAnswerFromBackend == null)
                {
                    //already handled in _RunTempTests
                    return;
                }

                foreach (var testResult in testResultAnswerFromBackend)
                {
                    var test = tests.FirstOrDefault(p => p.Id == testResult.TestId);

                    if (test == null)
                    {
                        await
                            Response.WriteAsync(
                                Jc.Serialize(new BasicResponse(ResponseCode.ServerError,
                                    "test not found for test result")));
                        return;
                    }

                    //get the old result
                    var oldResult = afterSolution.CustomTestResults
                        .FirstOrDefault(
                            p => p.CustomTestId == test.Id
                                 && p.AfterSolution == afterSolution);
                    //this is not a db query so the references are loaded

                    //and store test result

                    if (oldResult == null)
                    {
                        //no old result, create new one
                        oldResult = new CustomTestWithAfterSolutionAsTestResult()
                        {
                            AfterSolution = afterSolution,
                            CustomTest = test,
                            Passed = testResult.Passed,
                            Protocol = string.Join(Constants.TestProtocolSplitString, testResult.Protocol),
                            HasCompiled = testResult.HasCompiled,
                            ProgramExitCode = testResult.ProgramExitCode,
                            TestResultCode = testResult.TestResultCode,
                            RunnerVersion = testResult.RunnerVersion,
                            TestServerVersion= testResult.TestServerVersion,
                            TimeForCompiling = testResult.TimeForCompiling,
                            TimeForUserProgram = testResult.TimeForUserProgram,
                            TimeoutInMsUsed = testResult.TimeoutInMsUsed,
                            CompileTimeoutInMsUsed = testResult.CompileTimeoutInMsUsed,
                            CharacterLimitExceeded = testResult.CharacterLimitExceeded,
                            CharacterLimitUsed = testResult.CharacterLimitUsed,
                        };

                        _context.CustomTestWithAfterSolutionAsTestResults.Add(oldResult);
                    }
                    else
                    {
                        //change old result
                        oldResult.Passed = testResult.Passed;
                        oldResult.Protocol = string.Join(Constants.TestProtocolSplitString, testResult.Protocol);
                        oldResult.HasCompiled = testResult.HasCompiled;
                        oldResult.ProgramExitCode = testResult.ProgramExitCode;
                        oldResult.TestResultCode = testResult.TestResultCode;
                        
                        oldResult.RunnerVersion = testResult.RunnerVersion;
                        oldResult.TestServerVersion= testResult.TestServerVersion;
                        oldResult.TimeForCompiling = testResult.TimeForCompiling;
                        oldResult.TimeForUserProgram = testResult.TimeForUserProgram;
                        oldResult.TimeoutInMsUsed = testResult.TimeoutInMsUsed;
                        oldResult.CompileTimeoutInMsUsed = testResult.CompileTimeoutInMsUsed;
                        oldResult.CharacterLimitExceeded = testResult.CharacterLimitExceeded;
                        oldResult.CharacterLimitUsed = testResult.CharacterLimitUsed;
                    }

                    try
                    {
                        await _context.SaveChangesAsync();
                    }
                    catch (Exception ex)
                    {
                       await base.HandleDbError(ex);
                        return;
                    }
                }


                //user could request too many tests to one user blocks the tests server...
                //but some tests might be run...
                if (tooManyTestsRequested)
                {
                    if (false) //user can only run on the normal test server
                    {
                        await Response.WriteAsync(
                            Jc.Serialize(new BasicResponseWithData<List<TestAnswerFromBackend>>(ResponseCode.PartialOk,
                                "too many tests requested, max: "
                                + systemSettings.MaxNumberOfTestsWithOneRequestSubmitTestServer,
                                testResultAnswerFromBackend)));
                    }
                    else
                    {
                        //normal tests
                        await Response.WriteAsync(
                            Jc.Serialize(new BasicResponseWithData<List<TestAnswerFromBackend>>(ResponseCode.PartialOk,
                                "too many tests requested, max: "
                                + systemSettings.MaxNumberOfTestsWithOneRequest,
                                testResultAnswerFromBackend)));
                    }
                }
                else
                {
                    await Response.WriteAsync(
                        Jc.Serialize(new BasicResponseWithData<List<TestAnswerFromBackend>>(ResponseCode.Ok,
                            "tests ran",
                            testResultAnswerFromBackend)));
                }
            }
        }


        /// <summary>
        /// runs a custom test on the test server (called from RunTest)
        /// this does not checks the available working time or if the release is still released!
        /// </summary>
        /// <param name="runNormalTestCommand">the test to run (with all dependencies) from frontend</param>
        /// <param name="userId">the user id</param>
        /// <param name="plang">the plang to use</param>
        /// <param name="exerciseRelease">the exercise release to find the user solution</param>
        /// <param name="systemSettings"></param>
        private async Task _RunCustomTest(RunTestCommandForBackend runNormalTestCommand, int userId, PLang plang,
            ExerciseRelease exerciseRelease, SystemSetting systemSettings)
        {
            var participation = await _context.ExerciseReleaseWithUserAsParticipations
                .FirstOrDefaultAsync(
                p => p.ExerciseRelease.GeneratedCode == exerciseRelease.GeneratedCode && p.UserId == userId);

            if (participation == null)
            {
                return;
            }

            //check if available working time is up
            //this is already done in the calling method (where the http entry point is...)
            //if (ReleaseHelper.IsTimeUp(participation, exerciseRelease, this))
            //{
            //    return;
            //}


            //user is only allowed to run on the normal test server
            var testIds = runNormalTestCommand.TestIds.Take(systemSettings.MaxNumberOfTestsWithOneRequest)
                .ToList();
            ;

            var tests = await _context.CustomTest
                    //.Include(p => p.TestSettings)
                    .Include(p => p.TestType)
                    .Include(p => p.AssetReferences)
                    .ThenInclude(p => p.FileReferenceUserFileAsset)
                    .Where(p =>
                        p.UserId == userId &&
                        p.ExerciseReleaseId == exerciseRelease.Id &&
                        testIds.Contains(p.Id))
                    .ToListAsync()
                ;

            //frontend needs to partition all tests into chunks
            var tooManyTestsRequested = tests.Count != runNormalTestCommand.TestIds.Count;


            //do not include solution files here ... it might be too many data to load...
            //just find the correct version then start another query to get the solution version + files

            Solution userSolution = await _context.Solutions
                .Include(p => p.SolutionFiles)
                .Include(p =>
                    p.CustomTestResults) //only include custom tests here because its a custom test (not a normal test)
                .FirstOrDefaultAsync(
                    p => p.UserId == userId && p.ExerciseReleaseId == exerciseRelease.Id && p.PLangId == plang.Id);


            if (userSolution == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "solution not found")));
                return;
            }


            #region get default custom test settings

            //get default test case settings from exercise 


            DefaultCustomTestSettings defaultCustomTestSettings = await _context.DefaultCustomTestSettings
                .FirstOrDefaultAsync(p => p.ExerciseId == exerciseRelease.ExerciseId);

            if (defaultCustomTestSettings == null)
            {
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no custom test settings found")));
                return;
            }

            #endregion

            
            //get test asset contents
            var testFilesIdsToRead = tests.SelectMany(p => p.AssetReferences)
                .Select(p => p.FileReferenceUserFileAssetId)
                .ToArray();

            var testAssetsBasePath = Files.GetUploadFilePath(UploadDirType.UserAssets);

            var testFilesDict = await Files.ReadUploadedFilesAsArray(testAssetsBasePath, testFilesIdsToRead);


            List<AllTest> allTests;

            try
            {
                allTests = tests.Select(p => AllTest.CreateFrom(p, defaultCustomTestSettings, testFilesDict)).ToList();
            }
            catch (Exception e)
            {
                //we create this exception so it should be ok to send just the message...
                Console.WriteLine("ERROR " + e);
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, e.Message)));
                return;
            }


            var testResultAnswerFromBackend = await _RunTempTests(userId, userSolution, plang, allTests, systemSettings, false);

            if (testResultAnswerFromBackend == null)
            {
                //already handled in _RunTempTests
                return;
            }

            foreach (var testResult in testResultAnswerFromBackend)
            {
                var test = tests.FirstOrDefault(p => p.Id == testResult.TestId);

                //get the old custom test result
                var oldResult = userSolution.CustomTestResults
                    .FirstOrDefault(
                        p => p.CustomTestId == test.Id
                             && p.Solution == userSolution);
                //this is not a db query so the references are loaded


                //and store test result

                if (oldResult == null)
                {
                    //no old result, create new one
                    oldResult = new CustomTestWithSingleSolutionAsTestResult()
                    {
                        Solution = userSolution,
                        CustomTest = test,
                        Passed = testResult.Passed,
                        Protocol = string.Join(Constants.TestProtocolSplitString, testResult.Protocol),
                        HasCompiled = testResult.HasCompiled,
                        ProgramExitCode = testResult.ProgramExitCode,
                        TestResultCode = testResult.TestResultCode,
                        RunnerVersion = testResult.RunnerVersion,
                        TestServerVersion= testResult.TestServerVersion,
                        TimeForCompiling = testResult.TimeForCompiling,
                        TimeForUserProgram = testResult.TimeForUserProgram,
                        TimeoutInMsUsed = testResult.TimeoutInMsUsed,
                        CompileTimeoutInMsUsed = testResult.CompileTimeoutInMsUsed,
                        CharacterLimitExceeded = testResult.CharacterLimitExceeded,
                        CharacterLimitUsed = testResult.CharacterLimitUsed,
                    };

                    _context.CustomTestWithSingleSolutionAsTestResult.Add(oldResult);
                }
                else
                {
                    //change old result
                    oldResult.Passed = testResult.Passed;
                    oldResult.Protocol = string.Join(Constants.TestProtocolSplitString, testResult.Protocol);
                    oldResult.HasCompiled = testResult.HasCompiled;
                    oldResult.ProgramExitCode = testResult.ProgramExitCode;
                    oldResult.TestResultCode = testResult.TestResultCode;
                    
                    oldResult.RunnerVersion = testResult.RunnerVersion;
                    oldResult.TestServerVersion= testResult.TestServerVersion;
                    oldResult.TimeForCompiling = testResult.TimeForCompiling;
                    oldResult.TimeForUserProgram = testResult.TimeForUserProgram;
                    oldResult.TimeoutInMsUsed = testResult.TimeoutInMsUsed;
                    oldResult.CompileTimeoutInMsUsed = testResult.CompileTimeoutInMsUsed;
                    oldResult.CharacterLimitExceeded = testResult.CharacterLimitExceeded;
                    oldResult.CharacterLimitUsed = testResult.CharacterLimitUsed;
                }

                try
                {
                    await _context.SaveChangesAsync();
                }
                catch (Exception ex)
                {
                   await base.HandleDbError(ex);
                    return;
                }
            }

            //user could request too many tests to one user blocks the tests server...
            if (tooManyTestsRequested)
            {
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<List<TestAnswerFromBackend>>(ResponseCode.PartialOk,
                        "too many custom tests requested, max: "
                        + systemSettings.MaxNumberOfTestsWithOneRequest,
                        testResultAnswerFromBackend)));
            }
            else
            {
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<List<TestAnswerFromBackend>>(ResponseCode.Ok,
                        "custom test ran",
                        testResultAnswerFromBackend)));
            }
        }


        /// <summary>
        /// copied from _RunCustomTest
        /// </summary>
        [HttpPost("run/custom/project/custom")]
        public async Task RunCustomTestCustomProject(
            [FromBody] RunTestCommandCustomProjectForBackend runNormalTestCommand)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var systemSettings = await _context.SystemSettings.FirstOrDefaultAsync();

            if (systemSettings == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "system settings not found")));
                return;
            }
            
            PLang plang = await _context.PLangs.FirstOrDefaultAsync(p => p.Id == runNormalTestCommand.PlangId);

            if (plang == null)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "p lang not found")));
                return;
            }


            var customProject = await _context.CustomProjects
                    .FirstOrDefaultAsync(p => p.Id == runNormalTestCommand.CustomProjectId && p.UserId == userId)
                ;

            if (customProject == null)
            {
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project not found")));
                return;
            }

            var solution = await _context.CustomProjectSolutions
                    .Include(p => p.SolutionFiles)
                    .Include(p => p.TestResults)
                    .FirstOrDefaultAsync(p => p.CustomProjectId == customProject.Id && p.PLangId == plang.Id)
                ;

            if (solution == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "solution not found")));
                return;
            }


            var testIds = runNormalTestCommand.TestIds.Take(systemSettings.MaxNumberOfTestsWithOneRequest)
                .ToList();
            

            var tests = await _context.CustomProjectTests
                    .Include(p => p.CustomProjectTestSettings)
                    .Include(p => p.TestType)
                    .Include(p => p.AssetReferences)
                    .ThenInclude(p => p.FileReferenceUserFileAsset)
                    .Where(p => p.CustomProjectId == customProject.Id &&
                                testIds.Contains(p.Id))
                    .ToListAsync()
                ;

            var tooManyTestsRequested = tests.Count != runNormalTestCommand.TestIds.Count;
            
            //get test asset contents
            var testFilesIdsToRead = tests.SelectMany(p => p.AssetReferences)
                .Select(p => p.FileReferenceUserFileAssetId)
                .ToArray();

            var testAssetsBasePath = Files.GetUploadFilePath(UploadDirType.UserAssets);

            var testFilesDict = await Files.ReadUploadedFilesAsArray(testAssetsBasePath, testFilesIdsToRead);


            List<AllTest> allTests;

            try
            {
                allTests = tests.Select(p => AllTest.CreateFrom(p, testFilesDict)).ToList();
            }
            catch (Exception e)
            {
                //we create this exception so it should be ok to send just the message...
                Console.WriteLine("ERROR " + e);
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, e.Message)));
                return;
            }
            

            var solutionFiles = solution.SolutionFiles.Select(p => new SolutionFile()
            {
                Id = p.Id,
                Content = p.Content,
                DisplayIndex = p.DisplayIndex,
                FileNameWithExtension = p.FileNameWithExtension,
                IsDisplayed = p.IsDisplayed,
                TemplateFile = null,
                TemplateFileId = -1
            }).ToList();

            var mainFile = solutionFiles.FirstOrDefault(p => p.Id == solution.MainFileId);

            if (mainFile == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "main file not found")));
                return;
            }

            var tempSolution = new Solution()
            {
                CreatedAt = solution.CreatedAt,
                SolutionFiles = solutionFiles,
                CustomTestResults = null,
                Note = "",
                TestResults = null,
                MainFileId = solution.MainFileId,
                MainFile = mainFile
            };


            var testResultAnswerFromBackend = await _RunTempTests(userId, tempSolution, plang, allTests, systemSettings, false)
                ;

            if (testResultAnswerFromBackend == null)
            {
                //already handled in _RunTempTests
                return;
            }

            foreach (var testResult in testResultAnswerFromBackend)
            {
                var test = tests.FirstOrDefault(p => p.Id == testResult.TestId);


                //get the old custom test result
                var oldResult = solution.TestResults
                    .FirstOrDefault(
                        p => p.CustomProjectTestId == test.Id
                             && p.CustomProjectSolutionId == solution.Id);

                //and store test result

                if (oldResult == null)
                {
                    //no old result, create new one
                    oldResult = new CustomProjectTestWithSolutionAsTestResult()
                    {
                        CustomProjectSolution = solution,
                        CustomProjectTest = test,
                        Passed = testResult.Passed,
                        Protocol = string.Join(Constants.TestProtocolSplitString, testResult.Protocol),
                        HasCompiled = testResult.HasCompiled,
                        ProgramExitCode = testResult.ProgramExitCode,
                        TestResultCode = testResult.TestResultCode,
                        RunnerVersion = testResult.RunnerVersion,
                        TestServerVersion= testResult.TestServerVersion,
                        TimeForCompiling = testResult.TimeForCompiling,
                        TimeForUserProgram = testResult.TimeForUserProgram,
                        TimeoutInMsUsed = testResult.TimeoutInMsUsed,
                        CompileTimeoutInMsUsed = testResult.CompileTimeoutInMsUsed,
                        CharacterLimitExceeded = testResult.CharacterLimitExceeded,
                        CharacterLimitUsed = testResult.CharacterLimitUsed
                    };

                    _context.CustomProjectTestWithSolutionAsTestResults.Add(oldResult);
                }
                else
                {
                    //change old result
                    oldResult.Passed = testResult.Passed;
                    oldResult.Protocol = string.Join(Constants.TestProtocolSplitString, testResult.Protocol);
                    oldResult.HasCompiled = testResult.HasCompiled;
                    oldResult.ProgramExitCode = testResult.ProgramExitCode;
                    oldResult.TestResultCode = testResult.TestResultCode;

                    oldResult.RunnerVersion = testResult.RunnerVersion;
                    oldResult.TestServerVersion = testResult.TestServerVersion;
                    oldResult.TimeForCompiling = testResult.TimeForCompiling;
                    oldResult.TimeForUserProgram = testResult.TimeForUserProgram;
                    oldResult.TimeoutInMsUsed = testResult.TimeoutInMsUsed;
                    oldResult.CompileTimeoutInMsUsed = testResult.CompileTimeoutInMsUsed;
                    oldResult.CharacterLimitExceeded = testResult.CharacterLimitExceeded;
                    oldResult.CharacterLimitUsed = testResult.CharacterLimitUsed;
                }

                try
                {
                    await _context.SaveChangesAsync();
                }
                catch (Exception ex)
                {
                   await base.HandleDbError(ex);
                    return;
                }
            }

            //user could request too many tests to one user blocks the tests server...
            if (tooManyTestsRequested)
            {
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<List<TestAnswerFromBackend>>(ResponseCode.PartialOk,
                        "too many custom tests requested, max: "
                        + systemSettings.MaxNumberOfTestsWithOneRequest,
                        testResultAnswerFromBackend)));
            }
            else
            {
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<List<TestAnswerFromBackend>>(ResponseCode.Ok,
                        "custom test ran",
                        testResultAnswerFromBackend)));
            }
        }


        //--- tests where we NOT store the result

        /// <summary>
        /// runs a test as tutor (command contains solution to use)
        /// 
        /// tutor needs assess solutions to for the connected group to be allowed to run tests
        /// submit tests are run on the submit test server, normal tests are run on the normal test server
        ///
        /// NO results are stored! only redirected back to the frontend
        /// 
        /// ALSO make sure all solution files ha a unique id (to find the main file)
        ///
        /// files are used from frontend --> readonly replace is not need (replace solution readonly file with exercise template readonly file)
        /// </summary>
        /// <param name="tutorCommand"></param>
        [HttpPost("run/as/tutor")]
        public async Task RequestRunTestAsTutor([FromBody] RunTutorTestCommandForBackend tutorCommand)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            if (string.IsNullOrWhiteSpace(tutorCommand.ReleaseCode))
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release code not found")));
                return;
            }

            var exerciseRelease = await _context.ExerciseReleases
                    .Include(p => p.Exercise)
                    .FirstOrDefaultAsync(p => p.GeneratedCode == tutorCommand.ReleaseCode);

            if (exerciseRelease == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release not found")));
                return;
            }
            
            //allow this because running as tutor will not save the test result
//            if (exerciseRelease.Exercise == null)
//            {
//                await
//                    Response.WriteAsync(
//                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise not found")));
//                return;
//            }
//            
//            if (exerciseRelease.Exercise.IsPermanentlyLocked)
//            {
//                await
//                    Response.WriteAsync(
//                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "exercise is permanently locked")));
//                return;
//            }

            var userGroupId = await _context.ExerciseReleases
                .Where(p => p.GeneratedCode == tutorCommand.ReleaseCode)
                .Select(p => p.Exercise.UserGroupId)
                .FirstOrDefaultAsync();

            if (userGroupId <= 0)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "user group not found")));
                return;
            }

            if (await base.HasGroupPermission(userGroupId, (groupPermissions) => groupPermissions.CanAssessExercises) ==
                false)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no permission")));
                return;
            }


            if (tutorCommand.IsSubmitTest)
            {
                //should be only possible when the automatic assessment has finished
                if (exerciseRelease.HasAutomaticAssessmentFinished == false)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "automatic assessment has not finished yet")));
                    return;
                }
            }

            await _RequestRunAnyTestAsTutor(tutorCommand, exerciseRelease);
        }

        ///  <summary>
        ///  runs a test (normal test or submit test)
        ///  no checks are applied if the user has permission or the release is still active ... THIS NEEDS TO BE DONE IN THE CALLING METHOD
        ///  
        ///  submit tests are run on the submit test server, normal tests are run on the normal test server
        /// 
        ///  only checks for for solution (request) and user auth are applied
        /// 
        ///  </summary>
        ///  <param name="tutorCommand"></param>
        /// <param name="exerciseRelease"></param>
        private async Task _RequestRunAnyTestAsTutor(RunTutorTestCommandForBackend tutorCommand,
            ExerciseRelease exerciseRelease)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var systemSettings = await _context.SystemSettings.FirstOrDefaultAsync();

            if (systemSettings == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "system settings not found")));
                return;
            }
            
            if (!tutorCommand.IsNormalTest && !tutorCommand.IsSubmitTest)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "test type not set")));
                return;
            }

            //check if the solution has files... and the main file is valid

            if (tutorCommand.SolutionFiles.Count == 0)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "invalid solution")));
                return;
            }

            for (int i = 0; i < tutorCommand.SolutionFiles.Count; i++)
            {
                for (int j = i + 1; j < tutorCommand.SolutionFiles.Count; j++)
                {
                    if (tutorCommand.SolutionFiles[i].Id == tutorCommand.SolutionFiles[j].Id)
                    {
                        await
                            Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                "duplicate solution file id found")));
                        return;
                    }
                }
            }

            var files = new List<SolutionFile>();
            SolutionFile mainFile = null;

            for (int i = 0; i < tutorCommand.SolutionFiles.Count; i++)
            {
                var solutionFileForTestServer = tutorCommand.SolutionFiles[i];
                var solutionFile = new SolutionFile()
                {
                    Id = solutionFileForTestServer.Id,
                    Content = solutionFileForTestServer.Content,
                    FileNameWithExtension = solutionFileForTestServer.FileNameWithExtension,
                };
                files.Add(solutionFile);

                //check if the id is duplicated
                if (tutorCommand.MainFileId == solutionFile.Id)
                {
                    mainFile = solutionFile;
                }
            }

            if (mainFile == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "could not find main file")));
                return;
            }

            var tempSolution = new Solution()
            {
                CreatedAt = DateTime.Now,
                Note = "tutor test",
                SolutionFiles = files,
                MainFile = mainFile,
                MainFileId = mainFile.Id,
            };


            //then we can get the plang directly from the user solution...
            PLang pLangToUse = await _context.PLangs
                .FirstOrDefaultAsync(p => p.Id == tutorCommand.PlangId);

            if (pLangToUse == null)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "p lang not found")));
                return;
            }

            //get all tests

            #region get tests

            var testIds = tutorCommand.TestIds.Take(tutorCommand.IsSubmitTest
                    ? systemSettings.MaxNumberOfTestsWithOneRequestSubmitTestServer
                    : systemSettings.MaxNumberOfTestsWithOneRequest)
                .ToList();
            ;

            var tests = await _context.Tests
                    .Include(p => p.TestSettings)
                    .Include(p => p.TestType)
                    .Include(p => p.AssetReferences)
                    .ThenInclude(p => p.FileReferenceTestAsset)
                    .Where(p => p.IsSubmitTest == tutorCommand.IsSubmitTest &&
                                p.ExerciseId == exerciseRelease.ExerciseId &&
                                testIds.Contains(p.Id))
                    .ToListAsync()
                ;
            
            
            //get test asset contents
            var testFilesIdsToRead = tests.SelectMany(p => p.AssetReferences)
                .Select(p => p.FileReferenceTestAssetId)
                .ToArray();

            var testAssetsBasePath = Files.GetUploadFilePath(UploadDirType.TestAssets);

            var testFilesDict = await Files.ReadUploadedFilesAsArray(testAssetsBasePath, testFilesIdsToRead);

            List<AllTest> allTests;

            try
            {
                allTests = tests.Select(p => AllTest.CreateFrom(p, testFilesDict)).ToList();
            }
            catch (Exception e)
            {
                //we create this exception so it should be ok to send just the message...
                Console.WriteLine("ERROR " + e);
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, e.Message)));
                return;
            }

            var tooManyTestsRequested = tests.Count != tutorCommand.TestIds.Count;

            #endregion

            #region running tests

            var testResultAnswerFromBackend =
                    await _RunTempTests(userId, tempSolution, pLangToUse, allTests, systemSettings, tutorCommand.IsSubmitTest)
                ;

            if (testResultAnswerFromBackend == null)
            {
                //already handled in _RunTempTests
                return;
            }

            #region process test results

            foreach (var testResult in testResultAnswerFromBackend)
            {
                var test = tests.FirstOrDefault(p => p.Id == testResult.TestId);

                if (test == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.ServerError,
                                "test not found for test result")));
                    return;
                }
            }


            //user could request too many tests to one user blocks the tests server...
            //but some tests might be run...
            if (tooManyTestsRequested)
            {
                if (tutorCommand.IsSubmitTest)
                {
                    await Response.WriteAsync(
                        Jc.Serialize(new BasicResponseWithData<List<TestAnswerFromBackend>>(ResponseCode.PartialOk,
                            "too many tests requested, max: "
                            + systemSettings.MaxNumberOfTestsWithOneRequestSubmitTestServer,
                            testResultAnswerFromBackend)));
                }
                else
                {
                    //normal tests
                    await Response.WriteAsync(
                        Jc.Serialize(new BasicResponseWithData<List<TestAnswerFromBackend>>(ResponseCode.PartialOk,
                            "too many tests requested, max: "
                            + systemSettings.MaxNumberOfTestsWithOneRequest,
                            testResultAnswerFromBackend)));
                }
            }
            else
            {
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<List<TestAnswerFromBackend>>(ResponseCode.Ok, "tests ran",
                        testResultAnswerFromBackend)));
            }

            #endregion

            #endregion
        }


        //--- actual requests / some helpers

        /// <summary>
        /// runs a test and returns the results
        /// 
        /// at least one test result is returned (or null)
        /// </summary>
        /// <param name="userId">the user id to separate two requests</param>
        /// <param name="solution"></param>
        /// <param name="pLangToUse"></param>
        /// <param name="tests"></param>
        /// <param name="isSubmitTest">true: use submit test (timeout), false: normal test</param>
        private async Task<List<TestAnswerFromBackend>> _RunTempTests(int userId, Solution solution, PLang pLangToUse,
            List<AllTest> tests, SystemSetting systemSettings,  bool isSubmitTest = false)
        {
            bool tooManyTestsRequested = false;

            List<AllTest> allTests = new List<AllTest>();

            for (int i = 0; i < tests.Count; i++)
            {
                if (isSubmitTest)
                {
                    if (i >= systemSettings.MaxNumberOfTestsWithOneRequestSubmitTestServer)
                    {
                        tooManyTestsRequested = true;
                        break;
                    }
                }
                else
                {
                    //normal test
                    if (i >= systemSettings.MaxNumberOfTestsWithOneRequest)
                    {
                        tooManyTestsRequested = true;
                        break;
                    }
                }

                allTests.Add(tests[i]);
            }


            var allAssets = allTests.SelectMany(p => p.TestAssetsWithContent);
            List<ITestAsset> _testAssets = allAssets.ToList<ITestAsset>();
            if (!(await ValidateFileNames(solution.SolutionFiles, _testAssets))) return null;


            #region running tests

            Tuple<ResponseFromTestServer, string> answerFromTestServer = null;


            try
            {
                var testServerUrl = isSubmitTest ? systemSettings.SubmitTestServerUrl : systemSettings.TestServerUrl;
                //use the running userid as request distinct number, it's unlikely that the user runs two tests at the same time...
                answerFromTestServer = await _RunTestOnTestServer(solution, allTests, pLangToUse,
                    testServerUrl, isSubmitTest, userId, systemSettings);
            }
            catch (Exception ex)
            {
                Console.WriteLine("ERROR: test server timeout hit?: " + ex);
                //probably timeout hit (not from the test program but the request
                string reason = "";
                if (AppConfiguration.IsDebugMode)
                {
                    reason = ", error: " + ex;
                }

                await
                    Response.WriteAsync(
                        Jc.Serialize(
                            new BasicResponse(ResponseCode.ServerError, "could not reach test server" + reason)));
                return null;
            }

            if (answerFromTestServer?.Item1 == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError,
                            answerFromTestServer?.Item2)));
                return null;
            }

            //check if answer is correct

            if (answerFromTestServer.Item1.ResponseCode == (int) TestServerResponseCode.Ok)
            {
                //all ok but test could have failed
            }
            else if (answerFromTestServer.Item1.ResponseCode == (int) TestServerResponseCode.MaxParallelTestsReached)
            {
                //send the user answer to retry later
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.RetryLater, "")));
                return null;
            }
            else
            {
                //some other error
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError, "test server error code: "
                                                                                 +
                                                                                 answerFromTestServer.Item1
                                                                                     .ResponseCode + " message: " +
                                                                                 answerFromTestServer.Item1
                                                                                     .Message)));
                return null;
            }


            //safety check
            if (answerFromTestServer.Item1?.Data == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError, "invalid response from test server")));
                return null;
            }

            List<AnswerTestFromTestServer> testResults = answerFromTestServer.Item1.Data;

            if (testResults.Count == 0)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError, "invalid response from test server")));
                return null;
            }


            var testResultAnswerFromBackend = new List<TestAnswerFromBackend>();

            #region process test results

            foreach (var testResult in testResults)
            {
                var test = allTests.FirstOrDefault(p => p.Id == testResult.TestId);

                if (test == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.ServerError,
                                "test not found for test result")));
                    return null;
                }

                //in case the test-runner does not restrict this...
                bool characterLimitExceeded = testResult.Protocol.Length > YapexDbContext.TestProtocolMaxStringLength;

                if (characterLimitExceeded)
                {
                    //cut last part
                    testResult.Protocol = testResult.Protocol.Substring(0, YapexDbContext.TestProtocolMaxStringLength);
                    testResult.CharacterLimitExceeded = true;
                }
                
                string[] protocol =
                    testResult.Protocol.Split(new string[] {Constants.TestProtocolSplitString},
                        StringSplitOptions.RemoveEmptyEntries);

                //and give user feedback
                testResultAnswerFromBackend.Add(new TestAnswerFromBackend
                {
                    TestId = testResult.TestId,
                    Passed = testResult.Passed,
                    HasCompiled = testResult.HasCompiled,
                    ProgramExitCode = testResult.ProgramExitCode,
                    TestResultCode = testResult.TestResultCode,
                    Protocol = protocol.ToList(),
                    TestServerCode = testResult.TestServerCode,
                    TestServerMessage = testResult.TestServerMessage,
                    RunnerVersion = testResult.RunnerVersion,
                    TestServerVersion= testResult.TestServerVersion,
                    TimeForCompiling = testResult.TimeForCompiling,
                    TimeForUserProgram = testResult.TimeForUserProgram,
                    TimeoutInMsUsed = testResult.TimeoutInMsUsed,
                    CompileTimeoutInMsUsed = testResult.CompileTimeoutInMsUsed,
                    CharacterLimitExceeded = testResult.CharacterLimitExceeded,
                    CharacterLimitUsed = YapexDbContext.TestProtocolMaxStringLength
                });
            }


            //user could request too many tests to one user blocks the tests server...
            //but some tests might be run...
            if (tooManyTestsRequested)
            {
                //normal tests
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<List<TestAnswerFromBackend>>(ResponseCode.PartialOk,
                        "too many tests requested, max: "
                        + systemSettings.MaxNumberOfTestsWithOneRequest,
                        testResultAnswerFromBackend)));
            }


            return testResultAnswerFromBackend;

            #endregion

            #endregion
        }


        /// <summary>
        /// runs a test on the test server
        /// 
        /// <para></para>
        /// can throw error if e.g. timeout hit for the connection to the test server
        /// </summary>
        /// <param name="solution">the solution version to use</param>
        /// <param name="tests">the test to use</param>
        /// <param name="plang">the plang to use</param>
        /// <param name="testServerUrl">the url for the test server to use</param>
        /// <param name="runOnSubmitTestServer">true: use the submit test server timeout, false: use the normal test server timeout</param>
        /// <param name="userId">a number to help the test server to separate two requests that arrive at the same time
        /// </param>
        /// <returns>the answer from the test server </returns>
        internal async Task<Tuple<ResponseFromTestServer, string>> _RunTestOnTestServer(
            Solution solution, List<AllTest> tests,
            PLang plang, string testServerUrl, bool runOnSubmitTestServer, int userId, SystemSetting systemSettings)
        {
            var mainFile = solution.SolutionFiles.FirstOrDefault(p => p.Id == solution.MainFileId);

            if (mainFile == null)
            {
                Console.WriteLine("ERROR could not find main file in solution files... aborting");
                return new Tuple<ResponseFromTestServer, string>(null,
                    "could not find main file in solution files... aborting");
            }

            RunTestCommandForTestServer runTestCmd = null;


            //interlocked should be faster than a lock
            //also make a copy because requestSequentialNumber can change at any time
            long nextSequentialNumber = Interlocked.Increment(ref requestSequentialNumber);

            runTestCmd = new RunTestCommandForTestServer()
            {
                PLang = plang.InternalName,
                AllFiles = solution.SolutionFiles.Select(p => new SolutionFileForTestServer()
                {
                    FileName = p.FileNameWithExtension,
                    FileContent = p.Content
                }).ToList(),
                MainFileNameWithExtension = mainFile.FileNameWithExtension,
                Tests = tests.Select(test => new RunTestCommandDataForTestServer()
                {
                    TestId = test.Id,
                    TestContent = test.Content,
                    Command = test.TestType.InternalName,
                    TimeoutInMs = test.TestSettings.TimeoutInMs,
                    CompileTimeoutInMs = test.TestSettings.CompileTimeoutInMs,
                    MemoryLimitInKb = test.TestSettings.MemoryLimitInKb,
                    MaxDiskSpaceInKb = test.TestSettings.MaxDiskSpaceInKb,
                    CompilerOptions = test.TestSettings.CompilerOptions,
                    AllAssets = test.TestAssetsWithContent.Select(p => new TestAssetFileForTestServer()
                    {
                        FileName = p.DisplayName,
                        FileContent = p.Content,
                        Hash = p.Hash
                    }).ToList(),
                }).ToList(),
                RequestDistinctionString = userId + "_" + nextSequentialNumber,
                CharacterLimitProtocol = YapexDbContext.TestProtocolMaxStringLength
            };


            //run the test on the test server
            ResponseFromTestServer answerFromTestServer = null;

            var requestPayload = new StringContent(Jc.SerializePlain(runTestCmd), SubmissionController.UTF8WithoutBom,
                "application/json");


            HttpResponseMessage jsonResponse = null;
            //does not work because we use .result and this blocks... (or does it work??) if not tested... keep it
            var tokenSource = new CancellationTokenSource();

            bool errorDuringRequest = false;
            string errorMessageDuringRequest = "";

            using (var client = new HttpClient())
            {
                if (runOnSubmitTestServer)
                {
                    client.Timeout = new TimeSpan(0,0,0,0,systemSettings.SubmitTestServerTimeoutInMs);
                    tokenSource.CancelAfter(new TimeSpan(0,0,0,0,systemSettings.SubmitTestServerTimeoutInMs));
                }
                else
                {
                    client.Timeout = new TimeSpan(0,0,0,0,systemSettings.TestServerTimeoutInMs);
                    tokenSource.CancelAfter(new TimeSpan(0,0,0,0,systemSettings.TestServerTimeoutInMs));
                }

                try
                {
                    jsonResponse = await client.PostAsync(testServerUrl, requestPayload, tokenSource.Token);
                }
                catch (Exception ex)
                {
                    Console.WriteLine("ERROR getting test server result: " + ex);
                    //throw;
                    errorDuringRequest = true;

                    if (AppConfiguration.IsDebugMode)
                    {
                        errorMessageDuringRequest = "error during connection to test server: " + ex;
                    }
                    else
                    {
                        errorMessageDuringRequest = "error during connection to test server";
                    }                   
                }
            }

            string contentResult = null;

            if (jsonResponse != null)
                contentResult = await jsonResponse.Content.ReadAsStringAsync();
            
            
            tokenSource.Dispose();

            if (contentResult == null && tokenSource.IsCancellationRequested)
            {
                //test server timeout hit...
                return new Tuple<ResponseFromTestServer, string>(null,
                    "timeout waiting for test server response");
            }

            if (errorDuringRequest)
            {
                //e.g. not reachable
                return new Tuple<ResponseFromTestServer, string>(null,
                    errorMessageDuringRequest);
            }

            try
            {
                answerFromTestServer = Jc.Deserialize<ResponseFromTestServer>(contentResult);

                if (answerFromTestServer != null)
                {
                    if (answerFromTestServer.ResponseCode != (int) TestServerResponseCode.Ok && AppConfiguration.IsDebugMode)
                    {
                        Console.WriteLine("not ok response code from test server, content: " + contentResult);
                        return new Tuple<ResponseFromTestServer, string>(null, answerFromTestServer.Message);
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("ERROR deserializing, message: " + ex + ", time: " + DateTime.Now);
                Console.WriteLine("text tried to deserialize: " + contentResult);
                return new Tuple<ResponseFromTestServer, string>(null, "error deserializing test server response");;
            }


            return new Tuple<ResponseFromTestServer, string>(answerFromTestServer, contentResult);
        }
    }

    /// <summary>
    /// a test request/command for backend
    /// </summary>
    public class RunTestCommandForBackend
    {
        public RunTestCommandForBackend()
        {
            //should be a known test (kind)...
            Contract.Invariant(IsNormalTest || IsCustomTest || IsNormalTest);
        }

        /// <summary>
        /// the code from the exercise release
        /// </summary>
        public string GeneratedCode { get; set; }

        /// <summary>
        /// the p lang id to use 
        /// </summary>
        public int PlangId { get; set; }

        /// <summary>
        /// the test id... also ensure that the test is not a submit test
        /// </summary>
        public List<int> TestIds { get; set; }

        /// <summary>
        /// true: run a normal test (search for the test in the normal tests), false: not a normal test
        /// </summary>
        public bool IsNormalTest { get; set; }

        /// <summary>
        /// true: run a custom test (search for the test in the custom tests), false: not a custom test
        /// </summary>
        public bool IsCustomTest { get; set; }

        /// <summary>
        /// true: run a submit test (search for the test in the submit tests), false: not a submit test
        /// </summary>
        public bool IsSubmitTest { get; set; }
    }

    /// <summary>
    /// a test request/command for backend
    /// </summary>
    public class RunTestCommandCustomProjectForBackend
    {
        public int CustomProjectId { get; set; }

        /// <summary>
        /// the p lang id to use 
        /// </summary>
        public int PlangId { get; set; }

        /// <summary>
        /// the test id... also ensure that the test is not a submit test
        /// </summary>
        public List<int> TestIds { get; set; }
    }

    /// <summary>
    /// a test request/command for backend as a tutor (allow normal, submit tests)
    /// also the tutor tests have no saved solution but they are temporarily stored at frontend
    /// so the request contains the solution to use
    /// 
    /// a tutor needs to have the assess submissions group permission so we need the generated code
    /// </summary>
    public class RunTutorTestCommandForBackend
    {
        public RunTutorTestCommandForBackend()
        {
            //should be a known test (kind)...
            Contract.Invariant(IsNormalTest || IsSubmitTest);
        }

        /// <summary>
        /// the code from the exercise release
        /// </summary>
        public string ReleaseCode { get; set; }

        /// <summary>
        /// the p lang id to use 
        /// </summary>
        public int PlangId { get; set; }

        /// <summary>
        /// the test id... also ensure that the test is not a submit test
        /// </summary>
        public List<int> TestIds { get; set; }

        /// <summary>
        /// true: run a normal test (search for the test in the normal tests), false: not a normal test
        /// </summary>
        public bool IsNormalTest { get; set; }

        /// <summary>
        /// true: run a submit test (search for the test in the submit tests), false: not a submit test
        /// </summary>
        public bool IsSubmitTest { get; set; }

        /// <summary>
        /// the file to use as main file
        /// </summary>
        public int MainFileId { get; set; }

        /// <summary>
        /// the files
        /// </summary>
        public List<SolutionFileDoExerciseFullBase> SolutionFiles { get; set; }
    }


    /// <summary>
    /// a command to compile a single file in order to validate the syntax (this file is not saved in the db)
    /// </summary>
    public class RunCompileSingleFileCommandForBackend
    {
        /// <summary>
        /// the code from the exercise release (to check if the time is up)
        /// </summary>
        public string GeneratedCode { get; set; }

        /// <summary>
        /// the p lang id to use 
        /// </summary>
        public int PlangId { get; set; }

        /// <summary>
        /// the file to use as main file
        /// we cannot use an id because the file might not be saved yet
        /// </summary>
        public string MainFileName { get; set; }

        /// <summary>
        /// the files to compile
        /// </summary>
        public List<SolutionFileForTestServer> Files { get; set; }
    }

    /// <summary>
    /// a command to compile a single file in order to validate the syntax (this file is not saved in the db)
    /// </summary>
    public class RunCompileSingleFileCommandCustomProjectForBackend
    {
        /// <summary>
        /// the custom project (must be a project of the current user (so not every user can compile random)
        /// </summary>
        public int CustomProjectId { get; set; }

        /// <summary>
        /// the p lang id to use 
        /// </summary>
        public int PlangId { get; set; }

        /// <summary>
        /// the file to use as main file
        /// </summary>
        public string MainFileName { get; set; }

        /// <summary>
        /// the files to compile
        /// </summary>
        public List<SolutionFileForTestServer> Files { get; set; }
    }

    public class JustRunProgramTestCommandBase
    {
        /// <summary>
        /// the p lang id to use 
        /// </summary>
        public int PlangId { get; set; }

        /// <summary>
        /// the args to use
        /// </summary>
        public string CmdArgs { get; set; }
    }

    /// <summary>
    /// just run program command
    /// </summary>
    public class JustRunProgramTestCommandForBackend : JustRunProgramTestCommandBase
    {
        /// <summary>
        /// the code from the exercise release
        /// </summary>
        public string ReleaseCode { get; set; }
    }

    public class JustRunProgramTestAsTutorCommandForBackend : JustRunProgramTestCommandBase
    {
        /// <summary>
        /// the code from the exercise release
        /// </summary>
        public string ReleaseCode { get; set; }

        /// <summary>
        /// the p lang id to use 
        /// </summary>
        public int PlangId { get; set; }

        /// <summary>
        /// the file to use as main file
        /// we cannot use an id because the file might not be saved yet
        /// </summary>
        public string MainFileName { get; set; }

        /// <summary>
        /// the files to compile
        /// </summary>
        public List<SolutionFileForTestServer> Files { get; set; }
    }

    /// <summary>
    /// just run program command
    /// </summary>
    public class JustRunProgramTestCustomProjectCommandForBackend : JustRunProgramTestCommandBase
    {
        /// <summary>
        /// the code from the exercise release
        /// </summary>
        public int CustomProjectId { get; set; }
    }

    #region classes for communication with test server

    /// <summary>
    /// a run test command just for the test server (nothing for frontend)
    /// </summary>
    public class RunTestCommandForTestServer
    {
        /// <summary>
        /// this number is used in case the micro time is not enough to separate two requests 
        /// (because we need to create a dir with a different name for every run/request)
        /// so use this number to help separating two requests that arrive at the 'same' time...
        /// THIS MUST BE A STRING WITH ONLY letters, numbers, underscores because it may be used as dir/file names
        /// </summary>
        public string RequestDistinctionString { get; set; }

        /// <summary>
        /// the plang the solution files are written in
        /// </summary>
        public string PLang { get; set; }

        /// <summary>
        /// the main file name with extension
        /// </summary>
        public string MainFileNameWithExtension { get; set; }

        /// <summary>
        /// all solution files
        /// </summary>
        public List<SolutionFileForTestServer> AllFiles { get; set; }


        /// <summary>
        /// the tests to run
        /// </summary>
        public List<RunTestCommandDataForTestServer> Tests { get; set; }
        
        /// <summary>
        /// the db will only store a max of characters <see cref="YapexDbContext.TestProtocolMaxStringLength"/>
        /// for the test result protocol
        /// can be null for old results
        /// </summary>
        public int CharacterLimitProtocol { get; set; }
    }

    /// <summary>
    /// the tests with all data
    /// </summary>
    public class RunTestCommandDataForTestServer
    {
        /// <summary>
        /// the test id
        /// </summary>
        public int TestId { get; set; }

        /// <summary>
        /// the command to call
        /// </summary>
        public string Command { get; set; }

        /// <summary>
        /// the test content to use
        /// test server will throw if this is null
        /// </summary>
        public string TestContent { get; set; }

        /// <summary>
        /// the timeout in ms
        /// </summary>
        public int TimeoutInMs { get; set; }
        
        /// <summary>
        /// the compile timeout in ms
        /// </summary>
        public int CompileTimeoutInMs { get; set; }

        /// <summary>
        /// the memory limit in kb
        /// </summary>
        public int MemoryLimitInKb { get; set; }

        /// <summary>
        /// the max dispk space in kb the user program can write to
        /// </summary>
        public int MaxDiskSpaceInKb { get; set; }
        
        /// <summary>
        /// some compiler options .g. -Xlint options needs to be separated by whitespaces (like normal command args)
        /// </summary>
        public string CompilerOptions { get; set; }

        /// <summary>
        /// all test asset files
        /// </summary>
        public List<TestAssetFileForTestServer> AllAssets { get; set; }
    }

    /// <summary>
    /// a file for the test server
    /// </summary>
    public class SolutionFileForTestServer
    {
        /// <summary>
        /// the file name WITH extension
        /// </summary>
        public string FileName { get; set; }

        /// <summary>
        /// the file content
        /// </summary>
        public string FileContent { get; set; }
    }

    public class TestAssetFileForTestServer
    {
        public string FileName { get; set; }

        /// <summary>
        /// byte[] is sent via base64 encoding
        /// </summary>
        public byte[] FileContent { get; set; }

        /// <summary>
        /// the hash of the content
        /// </summary>
        public string Hash { get; set; }
    }


    public class ResponseFromTestServerBase
    {
        /// <summary>
        /// the response code <see cref="TestServerResponseCode"/>
        /// </summary>
        public int ResponseCode { get; set; }

        /// <summary>
        /// the message
        /// </summary>
        public string Message { get; set; }

        //Data is another prop here... maybe use generic?
    }

    /// <summary>
    /// the response from the test server
    /// </summary>
    public class ResponseFromTestServer : ResponseFromTestServerBase
    {
        /// <summary>
        /// the data (can be null on error)
        /// </summary>
        public List<AnswerTestFromTestServer> Data { get; set; }
    }

    /// <summary>
    /// a type base for the answer from the test server (and used also for sending back the result to the forntend)
    /// </summary>
    public class AnswerTestFromTestServerBase
    {
        /// <summary>
        /// the code for the test (see response codes from test server)
        /// </summary>
        public int TestServerCode { get; set; }

        /// <summary>
        /// some more information about the test if any error occurred
        /// </summary>
        public string TestServerMessage { get; set; }

        /// <summary>
        /// the test id
        /// </summary>
        public int TestId { get; set; }

        /// <summary>
        /// true: test was successful, false: not
        /// </summary>
        public bool Passed { get; set; }

        /// <summary>
        /// true: has compiled, fasle: not or if the test does not included a compile step this is also false
        /// </summary>
        public bool HasCompiled { get; set; }

        /// <summary>
        /// the exit code of the user program, not used yet (not event set)
        /// </summary>
        public int? ProgramExitCode { get; set; }

        /// <summary>
        /// the test result code (see test server codes)
        /// </summary>
        public int TestResultCode { get; set; }

        /// <summary>
        /// can be null
        /// </summary>
        public string RunnerVersion { get; set; }

        /// <summary>
        /// time for compiling in ms
        /// can be null if compiling was not finished or done
        /// </summary>
        public int? TimeForCompiling { get; set; }
        
        /// <summary>
        /// time needed for execution in ms
        /// can be null if execution was not finished or done
        /// </summary>
        public int? TimeForUserProgram { get; set; }
        
        /// <summary>
        /// null: not ran (e.g. because of test server error)
        /// else: the timeout in ms used for test execution
        /// </summary>
        public int? TimeoutInMsUsed { get; set; }
        
        /// <summary>
        /// null: not ran (e.g. because of test server error)
        /// else: the timeout in ms used for compiling only
        /// </summary>
        public int? CompileTimeoutInMsUsed { get; set; }
        
        /// <summary>
        /// the test server version
        /// can be null
        /// </summary>
        public string TestServerVersion { get; set; }
        
        /// <summary>
        /// true: the limit <see cref="YapexDbContext.TestProtocolMaxStringLength"/> was exceeded and
        /// the <see cref="Protocol"/> was cut
        /// </summary>
        public bool CharacterLimitExceeded { get; set; }
        /// <summary>
        /// the db will only store a max of characters <see cref="YapexDbContext.TestProtocolMaxStringLength"/>
        /// can be null for old results
        /// </summary>
        public int? CharacterLimitUsed { get; set; }
    }

    public class AnswerTestFromTestServer : AnswerTestFromTestServerBase
    {
        /// <summary>
        /// this is a list/array but json cannot decode multi line strings to split it at \r\n
        /// </summary>
        public string Protocol { get; set; } //[]
    }

    /// <summary>
    /// the result from the test (what we send back to the frontend)
    /// </summary>
    public class TestAnswerFromBackend : AnswerTestFromTestServerBase
    {
        /// <summary>
        /// the conversation protocol between the user program and the test runner
        /// </summary>
        public List<string> Protocol { get; set; }
    }

    #endregion
}
